__author__ = "dgilman"

# this file contains a class to compute lensing proprerties of a pseudo Navaro-Frenk-White profile with a core and truncation
# radius
import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from scipy.integrate import quad

__all__ = ["TNFWC"]


class TNFWC(LensProfileBase):
    """This class contains an pseudo NFW profile with a core radius and a truncation
    radius. The density in 3D is given by.

    .. math::
        \\rho(r) = \\frac{\\rho_0 r_s^3}{\\left(r^2+r_c^2\\right)^{1/2} \\left(r_s^2+r^2\\right)} \\left(\\frac{r_t^2}{r^2+r_t^2}\\right)

    When the core radius goes to zero and the truncation radius approaches infinity this profile reduces to an NFW profile
    with the squared term inside the parentheses.

    TODO: add the gravitational potential for this profile
    TODO: add analytic solution for 3D mass
    """

    profile_name = "TNFWC"
    param_names = ["Rs", "alpha_Rs", "center_x", "center_y", "r_trunc", "r_core"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "center_x": -100,
        "center_y": -100,
        "r_trunc": 0.001,
        "r_core": 0.00001,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "center_x": 100,
        "center_y": 100,
        "r_trunc": 1000.0,
        "r_core": 1000.0,
    }

    def derivatives(self, x, y, Rs, alpha_Rs, r_core, r_trunc, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function which are the deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection angle in x, deflection angle in y
        """
        rho0_input = self.alpha2rho0(alpha_Rs, Rs, r_core, r_trunc)
        Rs = np.maximum(Rs, 0.00000001)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        f_x, f_y = self.nfw_alpha(R, Rs, rho0_input, r_core, r_trunc, x_, y_)
        return f_x, f_y

    def hessian(self, x, y, Rs, alpha_Rs, r_core, r_trunc, center_x=0, center_y=0):
        """

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        rho0_input = self.alpha2rho0(alpha_Rs, Rs, r_core, r_trunc)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        kappa = self.density_2d(R, 0, Rs, rho0_input, r_core, r_trunc)
        gamma1, gamma2 = self.nfw_gamma(R, Rs, rho0_input, r_core, r_trunc, x_, y_)
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def density(R, Rs, rho0, r_core, r_trunc):
        """3D density profile.

        :param R: radius of interest
        :type Rs: scale radius
        :param rho0: central density normalization
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: rho(R) density
        """
        x = R / Rs
        beta = r_core / Rs
        tau = r_trunc / Rs
        denom_core = (beta**2 + x**2) ** 0.5
        denom_nfw = 1 + x**2
        denom_trunc = (x**2 + tau**2) / tau**2
        denom = denom_core * denom_nfw * denom_trunc
        return rho0 / denom

    def density_lens(self, r, Rs, alpha_Rs, r_core, r_trunc):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: 3d radios
        :param Rs: scale radius
        :param alpha_Rs: deflection at Rs
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: density rho(r)
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, r_core, r_trunc)
        return self.density(r, Rs, rho0, r_core, r_trunc)

    def density_2d(self, x, y, Rs, rho0, r_core, r_trunc, center_x=0, center_y=0):
        """2D (projected) density profile.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param rho0: density normalization at Rs
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :param center_x: profile center (same units as x)
        :param center_y: profile center (same units as x)
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        x = R / Rs
        beta = r_core / Rs
        tau = r_trunc / Rs
        Fx = self._f(x, beta, tau)
        return 2 * rho0 * Rs * Fx

    def mass_3d(self, r, Rs, rho0, r_core, r_trunc):
        """Mass enclosed a 3d sphere or radius r.

        :param r: 3d radius
        :param Rs: scale radius
        :param rho0: density normalization
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: M(<r)
        """
        integrand = lambda x: x**2 * self.density(x, Rs, rho0, r_core, r_trunc)
        return 4 * np.pi * quad(integrand, 0, r)[0]

    def mass_3d_lens(self, r, Rs, alpha_Rs, r_core, r_trunc):
        """Mass enclosed a 3d sphere or radius r. This function takes as input the
        lensing parameterization.

        :param r: 3d radius
        :param Rs: scale radius
        :param alpha_Rs: deflection angle at Rs
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: M(<r)
        """
        rho0 = self.alpha2rho0(alpha_Rs, Rs, r_core, r_trunc)
        m_3d = self.mass_3d(r, Rs, rho0, r_core, r_trunc)
        return m_3d

    def mass_2d(self, R, Rs, rho0, r_core, r_trunc):
        """Mass enclosed a 2d cylinder or projected radius R.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: mass in cylinder
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        beta = r_core / Rs
        tau = r_trunc / Rs
        gx = self._g(x, beta, tau)
        m_2d = 4 * rho0 * Rs * R**2 * gx / x**2 * np.pi
        return m_2d

    def nfw_alpha(self, R, Rs, rho0, r_core, r_trunc, ax_x, ax_y):
        """Deflection angle of the profile (times Sigma_crit D_OL) along the projection
        to coordinate 'axis'.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :param ax_x: x coordinate relative to center
        :param ax_y: y coordinate relative to center
        :return: Epsilon(R) projected density at radius R
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        beta = r_core / Rs
        tau = r_trunc / Rs
        gx = self._g(x, beta, tau)
        a = 4 * rho0 * Rs * R * gx / x**2 / R
        return a * ax_x, a * ax_y

    def nfw_gamma(self, R, Rs, rho0, r_core, r_trunc, ax_x, ax_y):
        """Shear gamma of NFW profile (times Sigma_crit) along the projection to
        coordinate 'axis'.

        :param R: 3d radius
        :param Rs: scale radius
        :param rho0: central density normalization
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :param ax_x: x coordinate relative to center
        :param ax_y: y coordinate relative to center
        :return: Epsilon(R) projected density at radius R
        """
        R = np.maximum(R, 0.00000001)
        x = R / Rs
        beta, tau = r_core / Rs, r_trunc / Rs
        gx = self._g(x, beta, tau)
        Fx = self._f(x, beta, tau)
        a = (
            2 * rho0 * Rs * (2 * gx / x**2 - Fx)
        )  # /x #2*rho0*Rs*(2*gx/x**2 - Fx)*axis/x
        return a * (ax_y**2 - ax_x**2) / R**2, -a * 2 * (ax_x * ax_y) / R**2

    def _f(self, x, b, t):
        """Analytic solution of the projection integral.

        :param X: R/Rs
        :type X: float >0
        :param b: core radius divided by the scale radius
        :param t: truncation radius divided by the scale radius
        :return: solution to the projection integral
        """
        prefactor = t**2 / (t**2 - 1)
        return prefactor * (self._u1(x, b, 1.0) - self._u1(x, b, t))

    def _g(self, x, b, t):
        """Analytic solution of integral for NFW profile to compute deflection angle and
        gamma.

        :param X: R/Rs
        :type X: float >0
        :param b: core radius divided by the scale radius
        :param t: truncation radius divided by the scale radius
        :return: solution of the integral over projected mass
        """
        if b == t:
            t += 1e-3
        prefactor = abs(t**2 / (t**2 - 1))
        return prefactor * (
            -self._u2(x, b, t)
            + self._u2(0.0, b, t)
            + self._u2(x, b, 1.0)
            - self._u2(0.0, b, 1.0)
        )

    @staticmethod
    def _u1(x, b, t):
        """
        :param x: R/Rs
        :param b: core radius divided by the scale radius
        :param t: truncation radius divided by the scale radius

        """
        t2x2 = t**2 + x**2
        b2x2 = b**2 + x**2
        b2mt2 = b**2 - t**2
        if t > b:
            func = np.arccosh
            b2mt2 *= -1
        else:
            func = np.arccos
        arg = np.sqrt(t2x2 / b2x2)
        return func(arg) / np.sqrt(t2x2 * b2mt2)

    def _u2(self, x, b, t):
        """
        :param x: R/Rs
        :param b: core radius divided by the scale radius
        :param t: truncation radius divided by the scale radius

        """
        return (t**2 + x**2) * self._u1(x, b, t)

    def alpha2rho0(self, alpha_Rs, Rs, r_core, r_trunc):
        """Convert angle at Rs into rho0.

        :param alpha_Rs: deflection angle at RS
        :param Rs: scale radius
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: density normalization
        """
        # alpha_Rs = 4 * rho0 * Rs * R * gx
        beta = r_core / Rs
        tau = r_trunc / Rs
        gx = self._g(1.0, beta, tau)
        rho0 = alpha_Rs / (4 * Rs**2 * gx)
        return rho0

    def rho02alpha(self, rho0, Rs, r_core, r_trunc):
        """Convert rho0 to angle at Rs.

        :param rho0: density normalization
        :param Rs: scale radius
        :param r_core: core radius [arcsec]
        :param r_trunc: truncation radius [arcsec]
        :return: deflection angle at RS
        """
        return self.nfw_alpha(Rs, Rs, rho0, r_core, r_trunc, Rs, 0.0)[0]



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.Analysis import light2mass
from lenstronomy.LightModel.light_model import LightModel


class TestLight2Mass(object):
    def setup_method(self):
        pass

    def test_light2mass_conversion(self):
        numPix = 100
        deltaPix = 0.05

        lightModel = LightModel(light_model_list=["SERSIC_ELLIPSE", "SERSIC"])
        kwargs_lens_light = [
            {"R_sersic": 0.5, "n_sersic": 4, "amp": 2, "e1": 0, "e2": 0.05},
            {"R_sersic": 1.5, "n_sersic": 1, "amp": 2},
        ]

        kwargs_interpol = light2mass.light2mass_interpol(
            lens_light_model_list=["SERSIC_ELLIPSE", "SERSIC"],
            kwargs_lens_light=kwargs_lens_light,
            numPix=numPix,
            deltaPix=deltaPix,
            subgrid_res=1,
        )
        from lenstronomy.LensModel.lens_model import LensModel

        lensModel = LensModel(lens_model_list=["INTERPOL_SCALED"])
        kwargs_lens = [kwargs_interpol]
        import lenstronomy.Util.util as util

        x_grid, y_grid = util.make_grid(numPix, deltapix=deltaPix)
        kappa = lensModel.kappa(x_grid, y_grid, kwargs=kwargs_lens)
        kappa = util.array2image(kappa)
        kappa /= np.mean(kappa)
        flux = lightModel.surface_brightness(x_grid, y_grid, kwargs_lens_light)
        flux = util.array2image(flux)
        flux /= np.mean(flux)
        # import matplotlib.pyplot as plt
        # plt.matshow(flux-kappa)
        # plt.colorbar()
        # plt.show()
        delta_kappa = (kappa - flux) / flux
        max_delta = np.max(np.abs(delta_kappa))
        assert max_delta < 1
        # assert max_diff < 0.01
        npt.assert_almost_equal(flux[0, 0], kappa[0, 0], decimal=2)


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy.testing as npt
import numpy as np
import unittest
import os

from lenstronomy.Util.coolest_interface import (
    create_lenstronomy_from_coolest,
    update_coolest_from_lenstronomy,
    create_kwargs_mcmc_from_chain_list,
)
from lenstronomy.Util.coolest_read_util import (
    degree_coolest_to_radian_lenstronomy,
    ellibounds_coolest_to_lenstronomy,
    shearbounds_coolest_to_lenstronomy,
)
from lenstronomy.Util.coolest_update_util import (
    shapelet_amp_lenstronomy_to_coolest,
    folding_coolest,
)

from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
import lenstronomy.Util.image_util as image_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Plots.model_plot import ModelPlot

from lenstronomy.Workflow.fitting_sequence import FittingSequence


class TestCOOLESTinterface(object):
    def test_load(self):
        path = os.getcwd()
        if path[-11:] == "lenstronomy":
            path += "/test/test_Util"
        kwargs_out = create_lenstronomy_from_coolest(path + "/coolest_template")
        print(kwargs_out)
        kwargs_out = create_lenstronomy_from_coolest(path + "/coolest_template_pemd")
        kwargs_out = create_lenstronomy_from_coolest(
            path + "/coolest_template_pemd_random"
        )
        return

    def test_update(self):
        path = os.getcwd()
        if path[-11:] == "lenstronomy":
            path += "/test/test_Util"
        kwargs_result = {
            "kwargs_lens": [
                {"gamma1": 0.1, "gamma2": -0.05, "ra_0": 0.0, "dec_0": 0.0},
                {
                    "theta_E": 0.7,
                    "e1": -0.15,
                    "e2": 0.01,
                    "center_x": 0.03,
                    "center_y": 0.01,
                },
            ],
            "kwargs_source": [
                {
                    "amp": 15.0,
                    "R_sersic": 0.11,
                    "n_sersic": 3.6,
                    "center_x": 0.02,
                    "center_y": -0.03,
                    "e1": 0.1,
                    "e2": -0.2,
                },
                {
                    "amp": np.array(
                        [70.0, 33.0, 2.1, 3.9, 15.0, -16.0, 2.8, -1.7, -4.1, 0.2]
                    ),
                    "n_max": 3,
                    "beta": 0.1,
                    "center_x": 0.1,
                    "center_y": 0.0,
                },
            ],
            "kwargs_lens_light": [
                {
                    "amp": 11.0,
                    "R_sersic": 0.2,
                    "n_sersic": 3.0,
                    "center_x": 0.03,
                    "center_y": 0.01,
                    "e1": -0.15,
                    "e2": 0.01,
                },
                {
                    "amp": 12.0,
                    "R_sersic": 0.02,
                    "n_sersic": 6.0,
                    "center_x": 0.03,
                    "center_y": 0.01,
                    "e1": 0.0,
                    "e2": -0.15,
                },
            ],
            "kwargs_ps": [
                {
                    "point_amp": np.array([0.1]),
                    "ra_image": np.array([0.25]),
                    "dec_image": np.array([0.2]),
                }
            ],
        }
        update_coolest_from_lenstronomy(
            path + "/coolest_template", kwargs_result, ending="_update"
        )
        kwargs_out = create_lenstronomy_from_coolest(path + "/coolest_template_update")
        npt.assert_almost_equal(
            kwargs_out["kwargs_params"]["lens_model"][0][1]["e1"],
            kwargs_result["kwargs_lens"][1]["e1"],
            decimal=4,
        )
        npt.assert_almost_equal(
            kwargs_out["kwargs_params"]["lens_model"][0][1]["e2"],
            kwargs_result["kwargs_lens"][1]["e2"],
            decimal=4,
        )

        return

    def test_full(self):
        # use read json ; create an image ; create noise ; do fit (PSO for result + MCMC for chain)
        # create the kwargs mcmc ; upadte json
        path = os.getcwd()
        if path[-11:] == "lenstronomy":
            path += "/test/test_Util"

        kwargs_out = create_lenstronomy_from_coolest(path + "/coolest_template")

        # IMAGE specifics
        background_rms = 0.005  # background noise per pixel
        exp_time = 500.0  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        # PSF : easier for test to create a gaussian PSF
        fwhm = 0.05  # full width at half maximum of PSF
        psf_type = "GAUSSIAN"  # 'GAUSSIAN', 'PIXEL', 'NONE'

        # lensing quantities to create an image
        lens_model_list = kwargs_out["kwargs_model"]["lens_model_list"]
        kwargs_sie = {
            "theta_E": 0.66,
            "center_x": 0.05,
            "center_y": 0,
            "e1": -0.1,
            "e2": 0.1,
        }  # parameters of the deflector lens model
        kwargs_shear = {
            "gamma1": 0.0,
            "gamma2": -0.05,
        }  # shear values to the source plane
        kwargs_lens = [kwargs_shear, kwargs_sie]
        lens_model_class = LensModel(lens_model_list)

        # Sersic parameters in the initial simulation for the source
        source_model_list = kwargs_out["kwargs_model"]["source_light_model_list"]
        kwargs_sersic = {
            "amp": 16,
            "R_sersic": 0.1,
            "n_sersic": 3.5,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.1,
            "center_y": 0,
        }
        kwargs_shapelets = {
            "amp": np.array([70.0, 33.0, 2.1, 3.9, 15.0, -16.0, 2.8, -1.7, -4.1, 0.2]),
            "n_max": 3,
            "beta": 0.1,
            "center_x": 0.1,
            "center_y": 0.0,
        }
        kwargs_source = [kwargs_sersic, kwargs_shapelets]
        source_model_class = LightModel(source_model_list)

        # Sersic parameters in the initial simulation for the lens light
        lens_light_model_list = kwargs_out["kwargs_model"]["lens_light_model_list"]
        kwargs_sersic_lens1 = {
            "amp": 16,
            "R_sersic": 0.6,
            "n_sersic": 2.5,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.05,
            "center_y": 0,
        }
        kwargs_sersic_lens2 = {
            "amp": 3,
            "R_sersic": 0.7,
            "n_sersic": 3,
            "e1": -0.1,
            "e2": 0.1,
            "center_x": 0.05,
            "center_y": 0,
        }
        kwargs_lens_light = [kwargs_sersic_lens1, kwargs_sersic_lens2]
        lens_light_model_class = LightModel(lens_light_model_list)

        numPix = 100
        kwargs_out["kwargs_data"]["background_rms"] = background_rms
        kwargs_out["kwargs_data"]["exposure_time"] = exp_time
        kwargs_out["kwargs_data"]["image_data"] = np.zeros((numPix, numPix))
        kwargs_out["kwargs_data"].pop("noise_map")

        data_class = ImageData(**kwargs_out["kwargs_data"])
        # PSF
        pixel_scale = (
            kwargs_out["kwargs_data"]["transform_pix2angle"][1][1]
            / kwargs_out["kwargs_psf"]["point_source_supersampling_factor"]
        )
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": pixel_scale,
            "truncation": 3,
        }
        kwargs_out["kwargs_psf"] = kwargs_psf
        psf_class = PSF(**kwargs_out["kwargs_psf"])

        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            kwargs_numerics=kwargs_numerics,
        )

        # generate image
        image_model = imageModel.image(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=None,
        )

        poisson = image_util.add_poisson(image_model, exp_time=exp_time)
        bkg = image_util.add_background(image_model, sigma_bkd=background_rms)
        image_real = image_model + poisson + bkg

        data_class.update_data(image_real)
        kwargs_out["kwargs_data"]["image_data"] = image_real

        # MODELING
        # Notes :
        # All the lines above were meant to create a mock image
        # The following is basically the only lines of code you will need
        # (after running the "create_lenstronomy_from_coolest" function) when you actually do the
        # modeling on a pre-existing image (with associated noise and psf proveded)
        band_list = [
            kwargs_out["kwargs_data"],
            kwargs_out["kwargs_psf"],
            kwargs_numerics,
        ]
        multi_band_list = [band_list]
        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "single-band",
        }
        kwargs_constraints = {
            "joint_lens_with_light": [
                [0, 1, ["center_x", "center_y", "e1", "e2"]],
                [1, 1, ["center_x", "center_y", "e1", "e2"]],
            ]
        }
        kwargs_likelihood = {"check_bounds": True, "check_positive_flux": True}

        fitting_seq = FittingSequence(
            kwargs_data_joint,
            kwargs_out["kwargs_model"],
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_out["kwargs_params"],
        )

        n_particules = 200
        n_iterations = 10
        wr = 5
        n_run_mcmc = 10
        n_burn_mcmc = 10
        fitting_kwargs_list = [
            [
                "PSO",
                {
                    "sigma_scale": 1.0,
                    "n_particles": n_particules,
                    "n_iterations": n_iterations,
                },
            ],
            [
                "MCMC",
                {
                    "n_burn": n_burn_mcmc,
                    "n_run": n_run_mcmc,
                    "walkerRatio": wr,
                    "sigma_scale": 0.01,
                },
            ],
        ]
        chain_list = fitting_seq.fit_sequence(fitting_kwargs_list)
        kwargs_result = fitting_seq.best_fit()

        modelPlot = ModelPlot(
            kwargs_data_joint["multi_band_list"],
            kwargs_out["kwargs_model"],
            kwargs_result,
        )
        modelPlot._imageModel.image_linear_solve(inv_bool=True, **kwargs_result)
        # the last 2 lines are meant for solving the linear parameters

        # use the function to save mcmc chains in userfriendly mode
        # kwargs_mcmc = create_kwargs_mcmc_from_chain_list(chain_list,kwargs_out['kwargs_model'],kwargs_out['kwargs_params'],
        #                                    kwargs_out['kwargs_data'],kwargs_out['kwargs_psf'],kwargs_numerics,
        #                                    kwargs_constraints,idx_chain=1)

        kwargs_mcmc = create_kwargs_mcmc_from_chain_list(
            chain_list,
            kwargs_out["kwargs_model"],
            kwargs_out["kwargs_params"],
            kwargs_out["kwargs_data"],
            kwargs_out["kwargs_psf"],
            kwargs_numerics,
            kwargs_constraints,
            idx_chain=1,
            likelihood_threshold=-100000,
        )
        # save the results (aka update the COOLEST json)
        update_coolest_from_lenstronomy(
            path + "/coolest_template", kwargs_result, kwargs_mcmc
        )

        return

    def test_pemd(self):
        path = os.getcwd()
        if path[-11:] == "lenstronomy":
            path += "/test/test_Util"
        kwargs_out = create_lenstronomy_from_coolest(path + "/coolest_template_pemd")
        print(kwargs_out)

        # kwargs_results to update the COOLEST template
        kwargs_result = {
            "kwargs_lens": [
                {
                    "theta_E": 0.7,
                    "e1": -0.15,
                    "e2": 0.01,
                    "gamma": 2.1,
                    "center_x": 0.03,
                    "center_y": 0.01,
                }
            ],
            "kwargs_source": [
                {
                    "amp": 15.0,
                    "R_sersic": 0.11,
                    "n_sersic": 3.6,
                    "center_x": 0.02,
                    "center_y": -0.03,
                    "e1": 0.1,
                    "e2": -0.2,
                }
            ],
            "kwargs_lens_light": [
                {
                    "amp": 11.0,
                    "R_sersic": 0.2,
                    "n_sersic": 3.0,
                    "center_x": 0.03,
                    "center_y": 0.01,
                    "e1": -0.15,
                    "e2": 0.01,
                }
            ],
        }
        # kwargs_mcmc to update the COOLEST template. In real cases, this list would be much bigger
        # as each element is a result from a given point at a given iteration of a MCMC chain

        kwargs_mcmc = {
            "args_lens": [
                [
                    {
                        "theta_E": 0.68,
                        "e1": -0.10,
                        "e2": -0.04,
                        "gamma": 1.9,
                        "center_x": 0.02,
                        "center_y": 0.10,
                    }
                ],
                [
                    {
                        "theta_E": 0.65,
                        "e1": -0.10,
                        "e2": -0.04,
                        "gamma": 1.9,
                        "center_x": 0.03,
                        "center_y": 0.01,
                    }
                ],
                [
                    {
                        "theta_E": 0.65,
                        "e1": -0.10,
                        "e2": -0.04,
                        "gamma": 1.9,
                        "center_x": 0.03,
                        "center_y": 0.01,
                    }
                ],
                [
                    {
                        "theta_E": 0.65,
                        "e1": -0.10,
                        "e2": -0.04,
                        "gamma": 1.9,
                        "center_x": 0.03,
                        "center_y": 0.01,
                    }
                ],
            ],
            "args_source": [
                [
                    {
                        "amp": 15.0,
                        "R_sersic": 0.11,
                        "n_sersic": 3.6,
                        "center_x": 0.02,
                        "center_y": -0.03,
                        "e1": 0.1,
                        "e2": -0.2,
                    }
                ],
                [
                    {
                        "amp": 15.0,
                        "R_sersic": 0.11,
                        "n_sersic": 3.6,
                        "center_x": 0.02,
                        "center_y": -0.03,
                        "e1": 0.1,
                        "e2": -0.2,
                    }
                ],
                [
                    {
                        "amp": 15.0,
                        "R_sersic": 0.11,
                        "n_sersic": 3.6,
                        "center_x": 0.02,
                        "center_y": -0.03,
                        "e1": 0.1,
                        "e2": -0.2,
                    }
                ],
                [
                    {
                        "amp": 15.0,
                        "R_sersic": 0.11,
                        "n_sersic": 3.6,
                        "center_x": 0.02,
                        "center_y": -0.03,
                        "e1": 0.1,
                        "e2": -0.2,
                    }
                ],
            ],
            "args_lens_light": [
                [
                    {
                        "amp": 11.0,
                        "R_sersic": 0.2,
                        "n_sersic": 3.0,
                        "center_x": 0.03,
                        "center_y": 0.01,
                        "e1": -0.15,
                        "e2": 0.01,
                    }
                ],
                [
                    {
                        "amp": 11.0,
                        "R_sersic": 0.2,
                        "n_sersic": 3.0,
                        "center_x": 0.03,
                        "center_y": 0.01,
                        "e1": -0.15,
                        "e2": 0.01,
                    }
                ],
                [
                    {
                        "amp": 11.0,
                        "R_sersic": 0.2,
                        "n_sersic": 3.0,
                        "center_x": 0.03,
                        "center_y": 0.01,
                        "e1": -0.15,
                        "e2": 0.01,
                    }
                ],
                [
                    {
                        "amp": 11.0,
                        "R_sersic": 0.2,
                        "n_sersic": 3.0,
                        "center_x": 0.03,
                        "center_y": 0.01,
                        "e1": -0.15,
                        "e2": 0.01,
                    }
                ],
            ],
        }
        update_coolest_from_lenstronomy(
            path + "/coolest_template_pemd",
            kwargs_result,
            ending="_update",
            kwargs_mcmc=kwargs_mcmc,
        )
        kwargs_out = create_lenstronomy_from_coolest(
            path + "/coolest_template_pemd_update"
        )
        print(kwargs_out)
        npt.assert_almost_equal(
            kwargs_out["kwargs_params"]["lens_model"][0][0]["e1"],
            kwargs_result["kwargs_lens"][0]["e1"],
            decimal=4,
        )
        npt.assert_almost_equal(
            kwargs_out["kwargs_params"]["lens_model"][0][0]["e2"],
            kwargs_result["kwargs_lens"][0]["e2"],
            decimal=4,
        )

        return

    def test_util_functions(self):
        radian = degree_coolest_to_radian_lenstronomy(None)
        radian = degree_coolest_to_radian_lenstronomy(-120.0)
        npt.assert_almost_equal(radian, np.pi / 6.0, decimal=4)

        radian = degree_coolest_to_radian_lenstronomy(120.0)
        npt.assert_almost_equal(radian, 5 * np.pi / 6.0, decimal=4)

        ellibounds_coolest_to_lenstronomy(0.6, 1.0, None, None)
        shearbounds_coolest_to_lenstronomy(0.0, 0.1, -90.0, None)

        shapelet_amp_lenstronomy_to_coolest(None)
        folding_coolest(np.array([-95.0, 95.0]))
        folding_coolest(-95.0)
        folding_coolest(95.0)

        return



import lenstronomy.Util.util as util
from lenstronomy.Util.param_util import shear_cartesian2polar
import lenstronomy.Util.simulation_util as sim_util
import matplotlib.pyplot as plt
from matplotlib import patches
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Profiles.curved_arc_spp import center_deflector
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Plots import plot_util
import scipy.ndimage as ndimage
from lenstronomy.Data.pixel_grid import PixelGrid

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


# TODO define coordinate grid beforehand, e.g. kwargs_data
# TODO feed in PointSource instance?


@export
def lens_model_plot(
    ax,
    lensModel,
    kwargs_lens,
    numPix=500,
    deltaPix=0.01,
    sourcePos_x=0,
    sourcePos_y=0,
    point_source=False,
    with_caustics=False,
    with_convergence=True,
    coord_center_ra=0,
    coord_center_dec=0,
    coord_inverse=False,
    fast_caustic=True,
    **kwargs
):
    """Plots a lens model (convergence) and the critical curves and caustics.

    :param ax: matplotlib axis instance
    :param lensModel: LensModel() class instance
    :param kwargs_lens: lens model keyword argument list
    :param numPix: total number of pixels (for convergence map)
    :param deltaPix: width of pixel (total frame size is deltaPix x numPix)
    :param sourcePos_x: float, x-position of point source (image positions computed by
        the lens equation)
    :param sourcePos_y: float, y-position of point source (image positions computed by
        the lens equation)
    :param point_source: bool, if True, illustrates and computes the image positions of
        the point source
    :param with_caustics: bool, if True, illustrates the critical curve and caustics of
        the system
    :param with_convergence: bool, if True, illustrates the convergence map
    :param coord_center_ra: float, x-coordinate of the center of the frame
    :param coord_center_dec: float, y-coordinate of the center of the frame
    :param coord_inverse: bool, if True, inverts the x-coordinates to go from right-to-
        left (effectively the RA definition)
    :param fast_caustic: boolean, if True, uses faster but less precise caustic
        calculation (might have troubles for the outer caustic (inner critical curve)
    :param with_convergence: boolean, if True, plots the convergence of the deflector
    :return:
    """
    kwargs_data = sim_util.data_configure_simple(
        numPix,
        deltaPix,
        center_ra=coord_center_ra,
        center_dec=coord_center_dec,
        inverse=coord_inverse,
    )
    data = ImageData(**kwargs_data)
    _coords = data
    _frame_size = numPix * deltaPix

    ra0, dec0 = data.radec_at_xy_0
    if coord_inverse:
        extent = [ra0, ra0 - _frame_size, dec0, dec0 + _frame_size]
    else:
        extent = [ra0, ra0 + _frame_size, dec0, dec0 + _frame_size]

    if with_convergence:
        kwargs_convergence = kwargs.get("kwargs_convergence", {})
        convergence_plot(
            ax,
            pixel_grid=_coords,
            lens_model=lensModel,
            kwargs_lens=kwargs_lens,
            extent=extent,
            **kwargs_convergence
        )
    if with_caustics is True:
        kwargs_caustics = kwargs.get("kwargs_caustics", {})
        caustics_plot(
            ax,
            pixel_grid=_coords,
            lens_model=lensModel,
            kwargs_lens=kwargs_lens,
            fast_caustic=fast_caustic,
            coord_inverse=coord_inverse,
            pixel_offset=True,
            **kwargs_caustics
        )
    if point_source:
        kwargs_point_source = kwargs.get("kwargs_point_source", {})
        point_source_plot(
            ax,
            pixel_grid=_coords,
            lens_model=lensModel,
            kwargs_lens=kwargs_lens,
            source_x=sourcePos_x,
            source_y=sourcePos_y,
            **kwargs_point_source
        )
    if coord_inverse:
        ax.set_xlim([ra0, ra0 - _frame_size])
    else:
        ax.set_xlim([ra0, ra0 + _frame_size])
    ax.set_ylim([dec0, dec0 + _frame_size])
    # ax.get_xaxis().set_visible(False)
    # ax.get_yaxis().set_visible(False)
    ax.autoscale(False)
    return ax


def convergence_plot(
    ax,
    pixel_grid,
    lens_model,
    kwargs_lens,
    extent=None,
    vmin=-1,
    vmax=1,
    cmap="Greys",
    **kwargs
):
    """Plot convergence.

    :param ax: matplotlib axis instance
    :param pixel_grid: lenstronomy PixelGrid() instance (or class with inheritance of
        PixelGrid()
    :param lens_model: LensModel() class instance
    :param kwargs_lens: lens model keyword argument list
    :param extent: [[min, max] [min, max]] of frame
    :param vmin: matplotlib vmin
    :param vmax: matplotlib vmax
    :param cmap: matplotlib cmap
    :param kwargs: keyword arguments for matshow
    :return: matplotlib axis instance with convergence plot
    """
    x_grid, y_grid = pixel_grid.pixel_coordinates
    x_grid1d = util.image2array(x_grid)
    y_grid1d = util.image2array(y_grid)
    kappa_result = lens_model.kappa(x_grid1d, y_grid1d, kwargs_lens)
    kappa_result = util.array2image(kappa_result)
    _ = ax.matshow(
        np.log10(kappa_result),
        origin="lower",
        extent=extent,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        **kwargs
    )
    return ax


def caustics_plot(
    ax,
    pixel_grid,
    lens_model,
    kwargs_lens,
    fast_caustic=True,
    coord_inverse=False,
    color_crit="r",
    color_caustic="g",
    pixel_offset=False,
    *args,
    **kwargs
):
    """

    :param ax: matplotlib axis instance
    :param pixel_grid: lenstronomy PixelGrid() instance (or class with inheritance of PixelGrid()
    :param lens_model: LensModel() class instance
    :param kwargs_lens: lens model keyword argument list
    :param fast_caustic: boolean, if True, uses faster but less precise caustic calculation
     (might have troubles for the outer caustic (inner critical curve)
    :param coord_inverse: bool, if True, inverts the x-coordinates to go from right-to-left
     (effectively the RA definition)
    :param color_crit: string, color of critical curve
    :param color_caustic: string, color of caustic curve
    :param pixel_offset: boolean; if True (default plotting), the coordinates are shifted a half a pixel to match with
     the matshow() command to center the coordinates in the pixel center
    :param args: argument for plotting curve
    :param kwargs: keyword arguments for plotting curves
    :return: updated matplotlib axis instance
    """
    lens_model_ext = LensModelExtensions(lens_model)
    pixel_width = pixel_grid.pixel_width
    frame_size = np.max(pixel_grid.width)
    coord_center_ra, coord_center_dec = pixel_grid.center
    ra0, dec0 = pixel_grid.radec_at_xy_0
    origin = [ra0, dec0]
    if fast_caustic:
        (
            ra_crit_list,
            dec_crit_list,
            ra_caustic_list,
            dec_caustic_list,
        ) = lens_model_ext.critical_curve_caustics(
            kwargs_lens,
            compute_window=frame_size,
            grid_scale=pixel_width,
            center_x=coord_center_ra,
            center_y=coord_center_dec,
        )
        points_only = False
    else:
        # only supports individual points due to output of critical_curve_tiling definition
        points_only = True
        ra_crit_list, dec_crit_list = lens_model_ext.critical_curve_tiling(
            kwargs_lens,
            compute_window=frame_size,
            start_scale=pixel_width,
            max_order=10,
            center_x=coord_center_ra,
            center_y=coord_center_dec,
        )
        ra_caustic_list, dec_caustic_list = lens_model.ray_shooting(
            ra_crit_list, dec_crit_list, kwargs_lens
        )
        # ra_crit_list, dec_crit_list = list(ra_crit_list), list(dec_crit_list)
        # ra_caustic_list, dec_caustic_list = list(ra_caustic_list), list(dec_caustic_list)
    plot_util.plot_line_set(
        ax,
        pixel_grid,
        ra_caustic_list,
        dec_caustic_list,
        color=color_caustic,
        origin=origin,
        flipped_x=coord_inverse,
        points_only=points_only,
        pixel_offset=pixel_offset,
        *args,
        **kwargs
    )
    plot_util.plot_line_set(
        ax,
        pixel_grid,
        ra_crit_list,
        dec_crit_list,
        color=color_crit,
        origin=origin,
        flipped_x=coord_inverse,
        points_only=points_only,
        pixel_offset=pixel_offset,
        *args,
        **kwargs
    )
    return ax


def point_source_plot(
    ax,
    pixel_grid,
    lens_model,
    kwargs_lens,
    source_x,
    source_y,
    name_list=None,
    **kwargs
):
    """Plots and illustrates images of a point source The plotting routine orders the
    image labels according to the arrival time and illustrates a diamond shape of the
    size of the magnification. The coordinates are chosen in pixel coordinates.

    :param ax: matplotlib axis instance
    :param pixel_grid: lenstronomy PixelGrid() instance (or class with inheritance of
        PixelGrid()
    :param lens_model: LensModel() class instance
    :param kwargs_lens: lens model keyword argument list
    :param source_x: x-position of source
    :param source_y: y-position of source
    :param name_list: list of names of images
    :type name_list: list of strings, longer or equal the number of point sources
    :param kwargs: additional plotting keyword arguments
    :return: matplotlib axis instance with figure
    """
    from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver

    solver = LensEquationSolver(lens_model)
    x_center, y_center = pixel_grid.center
    delta_pix = pixel_grid.pixel_width
    ra0, dec0 = pixel_grid.radec_at_xy_0
    tranform = pixel_grid.transform_angle2pix
    if (
        np.linalg.det(tranform) < 0
    ):  # if coordiate transform has negative parity (#TODO temporary fix)
        delta_pix_x = -delta_pix
    else:
        delta_pix_x = delta_pix
    origin = [ra0, dec0]

    theta_x, theta_y = solver.image_position_from_source(
        source_x,
        source_y,
        kwargs_lens,
        search_window=np.max(pixel_grid.width),
        x_center=x_center,
        y_center=y_center,
        min_distance=pixel_grid.pixel_width,
    )
    mag_images = lens_model.magnification(theta_x, theta_y, kwargs_lens)

    x_image, y_image = pixel_grid.map_coord2pix(theta_x, theta_y)
    if name_list is None:
        name_list = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K"]
    for i in range(len(x_image)):
        x_ = (x_image[i] + 0.5) * delta_pix_x + origin[0]
        y_ = (y_image[i] + 0.5) * delta_pix + origin[1]
        ax.plot(
            x_, y_, "dk", markersize=4 * (1 + np.log(np.abs(mag_images[i]))), alpha=0.5
        )
        ax.text(x_, y_, name_list[i], fontsize=20, color="k")
    x_source, y_source = pixel_grid.map_coord2pix(source_x, source_y)
    ax.plot(
        (x_source + 0.5) * delta_pix_x + origin[0],
        (y_source + 0.5) * delta_pix + origin[1],
        "*k",
        markersize=10,
    )
    return ax


@export
def arrival_time_surface(
    ax,
    lensModel,
    kwargs_lens,
    numPix=500,
    deltaPix=0.01,
    sourcePos_x=0,
    sourcePos_y=0,
    with_caustics=False,
    point_source=False,
    n_levels=10,
    kwargs_contours=None,
    image_color_list=None,
    letter_font_size=20,
    name_list=None,
):
    """

    :param ax: matplotlib axis instance
    :param lensModel: LensModel() class instance
    :param kwargs_lens: lens model keyword argument list
    :param numPix:
    :param deltaPix:
    :param sourcePos_x:
    :param sourcePos_y:
    :param with_caustics:
    :param point_source:
    :param name_list: list of names of images
    :type name_list: list of strings, longer or equal the number of point sources
    :return:
    """
    kwargs_data = sim_util.data_configure_simple(numPix, deltaPix)
    data = ImageData(**kwargs_data)
    ra0, dec0 = data.radec_at_xy_0
    origin = [ra0, dec0]
    _frame_size = numPix * deltaPix
    _coords = data
    x_grid, y_grid = data.pixel_coordinates
    lensModelExt = LensModelExtensions(lensModel)
    # ra_crit_list, dec_crit_list, ra_caustic_list, dec_caustic_list = lensModelExt.critical_curve_caustics(
    #    kwargs_lens, compute_window=_frame_size, grid_scale=deltaPix/2.)
    x_grid1d = util.image2array(x_grid)
    y_grid1d = util.image2array(y_grid)
    fermat_surface = lensModel.fermat_potential(
        x_grid1d, y_grid1d, kwargs_lens, sourcePos_x, sourcePos_y
    )
    fermat_surface = util.array2image(fermat_surface)
    if kwargs_contours is None:
        kwargs_contours = {}
        # , cmap='Greys', vmin=-1, vmax=1) #, cmap=self._cmap, vmin=v_min, vmax=v_max)
    if with_caustics is True:
        ra_crit_list, dec_crit_list = lensModelExt.critical_curve_tiling(
            kwargs_lens,
            compute_window=_frame_size,
            start_scale=deltaPix / 5,
            max_order=10,
        )
        ra_caustic_list, dec_caustic_list = lensModel.ray_shooting(
            ra_crit_list, dec_crit_list, kwargs_lens
        )
        plot_util.plot_line_set(
            ax, _coords, ra_caustic_list, dec_caustic_list, origin=origin, color="g"
        )
        plot_util.plot_line_set(
            ax, _coords, ra_crit_list, dec_crit_list, origin=origin, color="r"
        )
    if point_source is True:
        from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver

        solver = LensEquationSolver(lensModel)
        theta_x, theta_y = solver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=deltaPix,
            search_window=deltaPix * numPix,
        )

        fermat_pot_images = lensModel.fermat_potential(theta_x, theta_y, kwargs_lens)
        _ = ax.contour(
            x_grid,
            y_grid,
            fermat_surface,
            origin="lower",  # extent=[0, _frame_size, 0, _frame_size],
            levels=np.sort(fermat_pot_images),
            **kwargs_contours
        )
        # mag_images = lensModel.magnification(theta_x, theta_y, kwargs_lens)
        x_image, y_image = _coords.map_coord2pix(theta_x, theta_y)
        if name_list is None:
            name_list = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K"]

        for i in range(len(x_image)):
            x_ = (x_image[i] + 0.5) * deltaPix - _frame_size / 2
            y_ = (y_image[i] + 0.5) * deltaPix - _frame_size / 2
            if image_color_list is None:
                color = "k"
            else:
                color = image_color_list[i]
            ax.plot(x_, y_, "x", markersize=10, alpha=1, color=color)
            # markersize=8*(1 + np.log(np.abs(mag_images[i])))
            ax.text(
                x_ + deltaPix,
                y_ + deltaPix,
                name_list[i],
                fontsize=letter_font_size,
                color="k",
            )
        x_source, y_source = _coords.map_coord2pix(sourcePos_x, sourcePos_y)
        ax.plot(
            (x_source + 0.5) * deltaPix - _frame_size / 2,
            (y_source + 0.5) * deltaPix - _frame_size / 2,
            "*k",
            markersize=20,
        )
    else:
        vmin = np.min(fermat_surface)
        vmax = np.max(fermat_surface)
        levels = np.linspace(start=vmin, stop=vmax, num=n_levels)
        im = ax.contour(
            x_grid,
            y_grid,
            fermat_surface,
            origin="lower",  # extent=[0, _frame_size, 0, _frame_size],
            levels=levels,
            **kwargs_contours
        )
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.autoscale(False)
    return ax


@export
def curved_arc_illustration(
    ax, lensModel, kwargs_lens, with_centroid=True, stretch_scale=0.1, color="k"
):
    """

    :param ax: matplotlib axis instance
    :param lensModel: LensModel() instance
    :param kwargs_lens: list of lens model keyword arguments (only those of CURVED_ARC considered
    :param with_centroid: plots the center of the curvature radius
    :param stretch_scale: float, relative scale of banana to the tangential and radial stretches
     (effectively intrinsic source size)
    :param color: string, matplotlib color for plot
    :return: matplotlib axis instance
    """

    # loop through lens models
    # check whether curved arc
    lens_model_list = lensModel.lens_model_list
    for i, lens_type in enumerate(lens_model_list):
        if lens_type in [
            "CURVED_ARC",
            "CURVED_ARC_SIS_MST",
            "CURVED_ARC_CONST",
            "CURVED_ARC_CONST_MST",
            "CURVED_ARC_SPT",
            "CURVED_ARC_TAN_DIFF",
        ]:
            plot_arc(
                ax,
                with_centroid=with_centroid,
                stretch_scale=stretch_scale,
                color=color,
                **kwargs_lens[i]
            )

    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.autoscale(False)
    # rectangular frame
    ax.axis("scaled")

    # plot coordinate frame and scale


@export
def plot_arc(
    ax,
    tangential_stretch,
    radial_stretch,
    curvature,
    direction,
    center_x,
    center_y,
    stretch_scale=0.1,
    with_centroid=True,
    linewidth=1,
    color="k",
    dtan_dtan=0,
):
    """

    :param ax: matplotlib.axes instance
    :param tangential_stretch: float, stretch of intrinsic source in tangential direction
    :param radial_stretch: float, stretch of intrinsic source in radial direction
    :param curvature: 1/curvature radius
    :param direction: float, angle in radian
    :param center_x: center of source in image plane
    :param center_y: center of source in image plane
    :param with_centroid: plots the center of the curvature radius
    :param stretch_scale: float, relative scale of banana to the tangential and radial stretches
     (effectively intrinsic source size)
    :param linewidth: linewidth
    :param color: color
    :type color: string in matplotlib color convention
    :param dtan_dtan: tangential eigenvector differential in tangential direction (not implemented yet as illustration)
    :return:
    """
    # plot line to centroid
    center_x_spp, center_y_spp = center_deflector(
        curvature, direction, center_x, center_y
    )
    if with_centroid:
        ax.plot(
            [center_x, center_x_spp],
            [center_y, center_y_spp],
            "--",
            color=color,
            alpha=0.5,
            linewidth=linewidth,
        )
        ax.plot(
            [center_x_spp],
            [center_y_spp],
            "*",
            color=color,
            alpha=0.5,
            linewidth=linewidth,
        )

    # plot radial stretch to scale
    x_r = np.cos(direction) * radial_stretch * stretch_scale
    y_r = np.sin(direction) * radial_stretch * stretch_scale
    ax.plot(
        [center_x - x_r, center_x + x_r],
        [center_y - y_r, center_y + y_r],
        "--",
        color=color,
        linewidth=linewidth,
    )

    # compute angle of size of the tangential stretch
    r = 1.0 / curvature

    # make sure tangential stretch * stretch_scale is not larger than r * 2pi such that the full circle is only
    # plotted once
    tangential_stretch_ = min(tangential_stretch, np.pi * r / stretch_scale)
    d_phi = tangential_stretch_ * stretch_scale / r

    # linearly interpolate angle around center
    phi = np.linspace(-1, 1, 50) * d_phi + direction
    # plot points on circle
    x_curve = r * np.cos(phi) + center_x_spp
    y_curve = r * np.sin(phi) + center_y_spp
    ax.plot(x_curve, y_curve, "--", color=color, linewidth=linewidth)

    # make round circle with start point to end to close the circle
    r_c, t_c = util.points_on_circle(radius=stretch_scale, num_points=200)
    r_c = radial_stretch * r_c + r
    phi_c = t_c * tangential_stretch_ / r_c + direction
    x_c = r_c * np.cos(phi_c) + center_x_spp
    y_c = r_c * np.sin(phi_c) + center_y_spp
    ax.plot(x_c, y_c, "-", color=color, linewidth=linewidth)
    return ax

    # TODO add different colors for each quarter to identify parities


@export
def distortions(
    lensModel,
    kwargs_lens,
    num_pix=100,
    delta_pix=0.05,
    center_ra=0,
    center_dec=0,
    differential_scale=0.0001,
    smoothing_scale=None,
    **kwargs
):
    """

    :param lensModel: LensModel instance
    :param kwargs_lens: lens model keyword argument list
    :param num_pix: number of pixels per axis
    :param delta_pix: pixel scale per axis
    :param center_ra: center of the grid
    :param center_dec: center of the grid
    :param differential_scale: scale of the finite derivative length in units of angles
    :param smoothing_scale: float or None, Gaussian FWHM of a smoothing kernel applied before plotting
    :return: matplotlib instance with different panels
    """
    kwargs_grid = sim_util.data_configure_simple(
        num_pix, delta_pix, center_ra=center_ra, center_dec=center_dec
    )
    _coords = ImageData(**kwargs_grid)
    _frame_size = num_pix * delta_pix
    ra_grid, dec_grid = _coords.pixel_coordinates

    extensions = LensModelExtensions(lensModel=lensModel)
    ra_grid1d = util.image2array(ra_grid)
    dec_grid1d = util.image2array(dec_grid)
    (
        lambda_rad,
        lambda_tan,
        orientation_angle,
        dlambda_tan_dtan,
        dlambda_tan_drad,
        dlambda_rad_drad,
        dlambda_rad_dtan,
        dphi_tan_dtan,
        dphi_tan_drad,
        dphi_rad_drad,
        dphi_rad_dtan,
    ) = extensions.radial_tangential_differentials(
        ra_grid1d,
        dec_grid1d,
        kwargs_lens=kwargs_lens,
        center_x=center_ra,
        center_y=center_dec,
        smoothing_3rd=differential_scale,
        smoothing_2nd=None,
    )

    (
        lambda_rad2d,
        lambda_tan2d,
        orientation_angle2d,
        dlambda_tan_dtan2d,
        dlambda_tan_drad2d,
        dlambda_rad_drad2d,
        dlambda_rad_dtan2d,
        dphi_tan_dtan2d,
        dphi_tan_drad2d,
        dphi_rad_drad2d,
        dphi_rad_dtan2d,
    ) = (
        util.array2image(lambda_rad),
        util.array2image(lambda_tan),
        util.array2image(orientation_angle),
        util.array2image(dlambda_tan_dtan),
        util.array2image(dlambda_tan_drad),
        util.array2image(dlambda_rad_drad),
        util.array2image(dlambda_rad_dtan),
        util.array2image(dphi_tan_dtan),
        util.array2image(dphi_tan_drad),
        util.array2image(dphi_rad_drad),
        util.array2image(dphi_rad_dtan),
    )

    if smoothing_scale is not None:
        lambda_rad2d = ndimage.gaussian_filter(
            lambda_rad2d, sigma=smoothing_scale / delta_pix
        )
        dlambda_rad_drad2d = ndimage.gaussian_filter(
            dlambda_rad_drad2d, sigma=smoothing_scale / delta_pix
        )
        lambda_tan2d = np.abs(lambda_tan2d)
        # the magnification cut is made to make a stable integral/convolution
        lambda_tan2d[lambda_tan2d > 100] = 100
        lambda_tan2d = ndimage.gaussian_filter(
            lambda_tan2d, sigma=smoothing_scale / delta_pix
        )
        # the magnification cut is made to make a stable integral/convolution
        dlambda_tan_dtan2d[dlambda_tan_dtan2d > 100] = 100
        dlambda_tan_dtan2d[dlambda_tan_dtan2d < -100] = -100
        dlambda_tan_dtan2d = ndimage.gaussian_filter(
            dlambda_tan_dtan2d, sigma=smoothing_scale / delta_pix
        )
        orientation_angle2d = ndimage.gaussian_filter(
            orientation_angle2d, sigma=smoothing_scale / delta_pix
        )
        dphi_tan_dtan2d = ndimage.gaussian_filter(
            dphi_tan_dtan2d, sigma=smoothing_scale / delta_pix
        )

    def _plot_frame(ax, frame, vmin, vmax, text_string):
        """

        :param ax: matplotlib.axis instance
        :param frame: 2d array
        :param vmin: minimum plotting scale
        :param vmax: maximum plotting scale
        :param text_string: string to describe the label
        :return:
        """
        font_size = 10
        _arrow_size = 0.02
        im = ax.matshow(
            frame, extent=[0, _frame_size, 0, _frame_size], vmin=vmin, vmax=vmax
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax, orientation="vertical")
        # cb.set_label(text_string, fontsize=10)
        # plot_util.scale_bar(ax, _frame_size, dist=1, text='1"', font_size=font_size)
        plot_util.text_description(
            ax,
            _frame_size,
            text=text_string,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        # if 'no_arrow' not in kwargs or not kwargs['no_arrow']:
        #    plot_util.coordinate_arrows(ax, _frame_size, _coords,
        #                                color='w', arrow_size=_arrow_size,
        #                                font_size=font_size)

    f, axes = plt.subplots(3, 4, figsize=(12, 8))
    _plot_frame(
        axes[0, 0], lambda_rad2d, vmin=0.6, vmax=1.4, text_string=r"$\lambda_{rad}$"
    )
    _plot_frame(
        axes[0, 1], lambda_tan2d, vmin=-20, vmax=20, text_string=r"$\lambda_{tan}$"
    )
    _plot_frame(
        axes[0, 2],
        orientation_angle2d,
        vmin=-np.pi / 10,
        vmax=np.pi / 10,
        text_string=r"$\phi$",
    )
    _plot_frame(
        axes[0, 3],
        util.array2image(lambda_tan * lambda_rad),
        vmin=-20,
        vmax=20,
        text_string="magnification",
    )
    _plot_frame(
        axes[1, 0],
        dlambda_rad_drad2d / lambda_rad2d,
        vmin=-0.1,
        vmax=0.1,
        text_string="dlambda_rad_drad",
    )
    _plot_frame(
        axes[1, 1],
        dlambda_tan_dtan2d / lambda_tan2d,
        vmin=-20,
        vmax=20,
        text_string="dlambda_tan_dtan",
    )
    _plot_frame(
        axes[1, 2],
        dlambda_tan_drad2d / lambda_tan2d,
        vmin=-20,
        vmax=20,
        text_string="dlambda_tan_drad",
    )
    _plot_frame(
        axes[1, 3],
        dlambda_rad_dtan2d / lambda_rad2d,
        vmin=-0.1,
        vmax=0.1,
        text_string="dlambda_rad_dtan",
    )

    _plot_frame(
        axes[2, 0], dphi_rad_drad2d, vmin=-0.1, vmax=0.1, text_string="dphi_rad_drad"
    )
    _plot_frame(
        axes[2, 1],
        dphi_tan_dtan2d,
        vmin=0,
        vmax=20,
        text_string="dphi_tan_dtan: curvature radius",
    )
    _plot_frame(
        axes[2, 2], dphi_tan_drad2d, vmin=-0.1, vmax=0.1, text_string="dphi_tan_drad"
    )
    _plot_frame(
        axes[2, 3], dphi_rad_dtan2d, vmin=0, vmax=20, text_string="dphi_rad_dtan"
    )

    return f, axes


def stretch_plot(
    ax,
    lens_model,
    kwargs_lens,
    plot_grid=None,
    scale=1,
    ellipse_color="k",
    max_stretch=np.inf,
    **patch_kwargs
):
    """Plots ellipses at each point on a grid, scaled corresponding to the local
    Jacobian eigenvalues.

    :param ax: matplotib axis instance
    :param lens_model: LensModel instance
    :param kwargs_lens: lens model keyword argument list
    :param plot_grid: pixelgrid instance at which to draw ellipses. 'None' uses default.
    :param scale: scales sizes of drawn ellipses, bigger number=larger
    :param ellipse_color: color of ellipses, defaults to black
    :param max_stretch: optional max amount to stretch ellipses which sometimes diverge
    :param patch_kwargs: additional keyword arguments for creating ellipse patch
    :return: matplotlib axis instance with figure
    """

    if plot_grid is None:
        # define default ellipse grid (20x20 spanning from -2 to 2)
        plot_grid = PixelGrid(20, 20, np.array([[1, 0], [0, 1]]) * 0.2, -2, -2)
    lme = LensModelExtensions(lens_model)
    x_grid, y_grid = plot_grid.pixel_coordinates
    x = util.image2array(x_grid)
    y = util.image2array(y_grid)
    w1, w2, v11, v12, v21, v22 = lme.hessian_eigenvectors(x, y, kwargs_lens)
    stretch_1 = np.abs(1.0 / w1)  # stretch in direction of first eigenvalue (unsorted)
    stretch_2 = np.abs(1.0 / w2)
    stretch_direction = np.arctan2(
        v12, v11
    )  # Direction of first eigenvector. Other eigenvector is orthogonal.

    for i in range(len(stretch_direction)):
        stretch_1_amount = np.minimum(stretch_1[i], max_stretch)
        stretch_2_amount = np.minimum(stretch_2[i], max_stretch)
        ell = patches.Ellipse(
            (x[i], y[i]),
            stretch_1_amount * scale / 40,  # 40 arbitrarily chosen
            stretch_2_amount * scale / 40,
            angle=stretch_direction[i] * 180 / np.pi,
            linewidth=1,
            fill=False,
            color=ellipse_color,
            **patch_kwargs
        )
        ax.add_patch(ell)
    ax.set_xlim(np.min(x), np.max(x))
    ax.set_ylim(np.min(y), np.max(y))
    return ax


def shear_plot(
    ax,
    lens_model,
    kwargs_lens,
    plot_grid=None,
    scale=5,
    color="k",
    max_stretch=np.inf,
    **kwargs
):
    """Plots combined internal+external shear at each point on a grid, represented by
    pseudovectors in the direction of local shear with length corresponding to shear
    magnitude.

    :param ax: matplotib axis instance
    :param lens_model: LensModel instance
    :param kwargs_lens: lens model keyword argument list
    :param plot_grid: pixelgrid instance at which to draw pseudovectors
    :param scale: scales sizes of drawn pseudovectors, smaller number=larger vectors
    :param color: color of pseudovectors, defaults to black
    :param max_stretch: optional max amount to stretch ellipses which sometimes diverge
    :param kwargs: additional plotting keyword arguments
    :return: matplotlib axis instance with figure
    """

    if plot_grid is None:
        # define default ellipse grid (20x20 spanning from -2 to 2)
        plot_grid = PixelGrid(20, 20, np.array([[1, 0], [0, 1]]) * 0.2, -2, -2)

    x_grid, y_grid = plot_grid.pixel_coordinates
    g1, g2 = lens_model.gamma(x_grid, y_grid, kwargs_lens)
    phi, shear = shear_cartesian2polar(g1, g2)
    max_stretch_array = np.ones_like(shear) * max_stretch
    shear = np.minimum(shear, max_stretch_array)
    arrow_x = shear * np.cos(phi)
    arrow_y = shear * np.sin(phi)
    ax.quiver(
        x_grid,
        y_grid,
        arrow_x,
        arrow_y,
        headaxislength=0,
        headlength=0,
        pivot="middle",
        scale=scale,
        linewidth=0.5,
        units="xy",
        width=0.02,
        headwidth=1,
        color=color,
        **kwargs
    )
    # , headwidth=0, headlength=0)
    ax.set_xlim(np.min(x_grid), np.max(x_grid))
    ax.set_ylim(np.min(y_grid), np.max(y_grid))
    return ax



from lenstronomy.LensModel import convergence_integrals
import lenstronomy.Util.util as util
from lenstronomy.LensModel.Profiles.sis import SIS
import numpy.testing as npt
import pytest


class TestConvergenceIntegrals(object):
    """Test angular to mass unit conversions."""

    def setup_method(self):
        pass

    def test_potential_from_kappa(self):
        sis = SIS()
        deltaPix = 0.005
        x_grid, y_grid = util.make_grid(numPix=2000, deltapix=deltaPix)
        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}

        f_xx, _, _, f_yy = sis.hessian(x_grid, y_grid, **kwargs_sis)
        f_ = sis.function(x_grid, y_grid, **kwargs_sis)
        f_ = util.array2image(f_)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        potential_num = convergence_integrals.potential_from_kappa_grid(kappa, deltaPix)

        x1, y1 = 560, 500
        x2, y2 = 550, 500
        # test relative potential at two different point way inside the kappa map
        d_f_num = potential_num[x1, y1] - potential_num[x2, y2]
        d_f = f_[x1, y1] - f_[x2, y2]
        npt.assert_almost_equal(d_f_num, d_f, decimal=2)

    def test_potential_from_kappa_adaptiv(self):
        sis = SIS()
        deltaPix = 0.01
        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        low_res_factor = 5
        high_res_kernel_size = 5
        x_grid, y_grid = util.make_grid(numPix=1000, deltapix=deltaPix)

        f_xx, _, _, f_yy = sis.hessian(x_grid, y_grid, **kwargs_sis)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_num = convergence_integrals.potential_from_kappa_grid_adaptive(
            kappa, deltaPix, low_res_factor, high_res_kernel_size
        )

        x_grid_low, y_grid_low = util.make_grid(
            numPix=1000 / low_res_factor, deltapix=deltaPix * low_res_factor
        )
        f_low = sis.function(x_grid_low, y_grid_low, **kwargs_sis)
        f_low = util.array2image(f_low)
        x1, y1 = 56, 50
        x2, y2 = 55, 50
        # test relative potential at two different point way inside the kappa map
        d_f_num = f_num[x1, y1] - f_num[x2, y2]
        d_f = f_low[x1, y1] - f_low[x2, y2]
        npt.assert_almost_equal(d_f_num, d_f, decimal=2)

    def test_deflection_from_kappa(self):
        sis = SIS()
        deltaPix = 0.01
        x_grid, y_grid = util.make_grid(numPix=1000, deltapix=deltaPix)
        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}

        f_xx, _, _, f_yy = sis.hessian(x_grid, y_grid, **kwargs_sis)
        f_x, f_y = sis.derivatives(x_grid, y_grid, **kwargs_sis)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )

        x1, y1 = 550, 500
        # test relative potential at two different point way inside the kappa map
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)

    def test_deflection_from_kappa_adaptiv(self):
        sis = SIS()
        deltaPix = 0.01
        kwargs_sis = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        low_res_factor = 5
        high_res_kernel_size = 5
        x_grid, y_grid = util.make_grid(numPix=1000, deltapix=deltaPix)

        f_xx, _, _, f_yy = sis.hessian(x_grid, y_grid, **kwargs_sis)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid_adaptive(
            kappa, deltaPix, low_res_factor, high_res_kernel_size
        )

        x_grid_low, y_grid_low = util.make_grid(
            numPix=1000 / low_res_factor, deltapix=deltaPix * low_res_factor
        )
        f_x_low, f_y_low = sis.derivatives(x_grid_low, y_grid_low, **kwargs_sis)
        f_x_low = util.array2image(f_x_low)
        f_y_low = util.array2image(f_y_low)
        x1, y1 = 50, 51
        # test relative potential at two different point way inside the kappa map
        npt.assert_almost_equal(f_x_low[x1, y1], f_x_num[x1, y1], decimal=2)
        print(f_x_low[x1, y1], f_x_num[x1, y1])
        npt.assert_almost_equal(f_y_low[x1, y1], f_y_num[x1, y1], decimal=2)

    def test_sersic(self):
        from lenstronomy.LensModel.Profiles.sersic import Sersic
        from lenstronomy.LightModel.Profiles.sersic import Sersic as SersicLight

        sersic_lens = Sersic()
        sersic_light = SersicLight()
        kwargs_light = {
            "n_sersic": 2,
            "R_sersic": 0.5,
            "I0_sersic": 1,
            "center_x": 0,
            "center_y": 0,
        }
        kwargs_lens = {
            "n_sersic": 2,
            "R_sersic": 0.5,
            "k_eff": 1,
            "center_x": 0,
            "center_y": 0,
        }
        deltaPix = 0.01
        numPix = 1000
        x_grid, y_grid = util.make_grid(numPix=numPix, deltapix=deltaPix)
        x_grid2d = util.array2image(x_grid)
        y_grid2d = util.array2image(y_grid)

        f_xx, _, _, f_yy = sersic_lens.hessian(x_grid, y_grid, **kwargs_lens)
        f_x, f_y = sersic_lens.derivatives(x_grid, y_grid, **kwargs_lens)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )
        x1, y1 = 500, 550
        x0, y0 = int(numPix / 2.0), int(numPix / 2.0)
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)
        f_num = convergence_integrals.potential_from_kappa_grid(kappa, deltaPix)
        f_ = sersic_lens.function(x_grid2d[x1, y1], y_grid2d[x1, y1], **kwargs_lens)
        f_00 = sersic_lens.function(x_grid2d[x0, y0], y_grid2d[x0, y0], **kwargs_lens)
        npt.assert_almost_equal(f_ - f_00, f_num[x1, y1] - f_num[x0, y0], decimal=2)

    def test_gnfw(self):
        from lenstronomy.LensModel.Profiles.general_nfw import GNFW

        gnfw = GNFW()
        deltaPix = 0.005
        numPix = 2000
        x_grid, y_grid = util.make_grid(numPix=numPix, deltapix=deltaPix)

        kwargs_lens = {
            "alpha_Rs": 1.2,
            "Rs": 0.8,
            "gamma_inner": 1.5,
            "gamma_outer": 3.4,
        }
        f_xx, _, _, f_yy = gnfw.hessian(x_grid, y_grid, **kwargs_lens)
        f_x, f_y = gnfw.derivatives(x_grid, y_grid, **kwargs_lens)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )
        x1, y1 = 500, 550
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)

        kwargs_lens = {
            "alpha_Rs": 1.2,
            "Rs": 0.8,
            "gamma_inner": 2.3,
            "gamma_outer": 3.45,
        }
        f_xx, _, _, f_yy = gnfw.hessian(x_grid, y_grid, **kwargs_lens)
        f_x, f_y = gnfw.derivatives(x_grid, y_grid, **kwargs_lens)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )
        x1, y1 = 500, 550
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)

    def test_tnfwc(self):
        from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC

        tnfwc = TNFWC()
        deltaPix = 0.005
        numPix = 2000
        x_grid, y_grid = util.make_grid(numPix=numPix, deltapix=deltaPix)

        kwargs_lens = {"alpha_Rs": 1.2, "Rs": 0.8, "r_trunc": 3.5, "r_core": 0.45}
        f_xx, _, _, f_yy = tnfwc.hessian(x_grid, y_grid, **kwargs_lens)
        f_x, f_y = tnfwc.derivatives(x_grid, y_grid, **kwargs_lens)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )
        x1, y1 = 500, 550
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)

        kwargs_lens = {"alpha_Rs": 1.2, "Rs": 0.8, "r_trunc": 300.5, "r_core": 0.45}
        f_xx, _, _, f_yy = tnfwc.hessian(x_grid, y_grid, **kwargs_lens)
        f_x, f_y = tnfwc.derivatives(x_grid, y_grid, **kwargs_lens)
        f_x = util.array2image(f_x)
        kappa = util.array2image((f_xx + f_yy) / 2.0)
        f_x_num, f_y_num = convergence_integrals.deflection_from_kappa_grid(
            kappa, deltaPix
        )
        x1, y1 = 500, 550
        npt.assert_almost_equal(f_x[x1, y1], f_x_num[x1, y1], decimal=2)


if __name__ == "__main__":
    pytest.main()



"""This module contains a class to compute the elliptical Navarro-Frank-White function
in mass/kappa space."""

__author__ = "ajshajib"

from lenstronomy.LensModel.Profiles.nfw_mass_concentration import NFWMC
from lenstronomy.LensModel.Profiles.nfw_ellipse_cse import NFW_ELLIPSE_CSE


class NFWMCEllipse(NFWMC):
    """This class contains functions parameterises the NFW profile with log10 M200 and
    the concentration rs/r200 relation are: R_200 = c * Rs.

    ATTENTION: the parameterization is cosmology and redshift dependent!
    The cosmology to connect mass and deflection relations is fixed to default H0=70km/s
    Omega_m=0.3 flat LCDM. It is recommended to keep a given cosmology definition in the
    lens modeling as the observable reduced deflection angles are sensitive in this
    parameterization. If you do not want to impose a mass-concentration relation, it is
    recommended to use the default NFW lensing profile parameterized in reduced
    deflection angles.
    """

    param_names = ["logM", "concentration", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "logM": 0,
        "concentration": 0.01,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "logM": 16,
        "concentration": 1000,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, z_lens, z_source, cosmo=None, static=False):
        """

        :param z_lens: redshift of lens
        :param z_source: redshift of source
        :param cosmo: astropy cosmology instance
        :param static: boolean, if True, only operates with fixed parameter values
        """
        super(NFWMCEllipse, self).__init__(z_lens, z_source, cosmo, static)
        self._nfw = NFW_ELLIPSE_CSE()

    def function(self, x, y, logM, concentration, e1, e2, center_x=0, center_y=0):
        """Compute the lensing potential of the NFW profile with ellipticity.

        :param x: angular position
        :param y: angular position
        :param Rs: angular turn over point
        :param alpha_Rs: deflection at Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo
        :param center_y: center of halo
        :return:
        """
        Rs, alpha_Rs = self._m_c2deflections(logM, concentration)
        return self._nfw.function(
            x,
            y,
            alpha_Rs=alpha_Rs,
            Rs=Rs,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )

    def derivatives(self, x, y, logM, concentration, e1, e2, center_x=0, center_y=0):
        """Return df/dx and df/dy of the function (integral of NFW).

        :param x: angular position
        :param y: angular position
        :param Rs: angular turn over point
        :param alpha_Rs: deflection at Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo
        :param center_y: center of halo
        :return:
        """
        Rs, alpha_Rs = self._m_c2deflections(logM, concentration)
        return self._nfw.derivatives(x, y, Rs, alpha_Rs, e1, e2, center_x, center_y)

    def hessian(self, x, y, logM, concentration, e1, e2, center_x=0, center_y=0):
        """Return Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2.

        :param x: angular position
        :param y: angular position
        :param Rs: angular turn over point
        :param alpha_Rs: deflection at Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo
        :param center_y: center of halo
        :return:
        """
        Rs, alpha_Rs = self._m_c2deflections(logM, concentration)
        return self._nfw.hessian(x, y, Rs, alpha_Rs, e1, e2, center_x, center_y)



import numpy as np
from lenstronomy.LensModel.QuadOptimizer.multi_plane_fast import MultiplaneFast
from lenstronomy.LensModel.QuadOptimizer.param_manager import PowerLawFreeShear
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
import numpy.testing as npt
import pytest


class TestFastRayShooting(object):
    def setup_method(self):
        self.zlens, self.zsource = 0.5, 1.5
        epl_kwargs = {
            "theta_E": 0.8,
            "center_x": 0.1,
            "center_y": 0.0,
            "e1": -0.2,
            "e2": 0.1,
            "gamma": 2.05,
        }
        shear_kwargs = {"gamma1": 0.09, "gamma2": -0.02}
        kwargs_macro = [epl_kwargs, shear_kwargs]

        self.x_image = np.array([0.65043538, -0.31109505, 0.78906059, -0.86222271])
        self.y_image = np.array([-0.89067493, 0.94851787, 0.52882605, -0.25403778])

        halo_list = ["SIS", "SIS", "SIS"]
        halo_z = [self.zlens - 0.1, self.zlens, self.zlens + 0.4]
        halo_kwargs = [
            {"theta_E": 0.05, "center_x": 0.3, "center_y": -0.9},
            {"theta_E": 0.01, "center_x": 1.3, "center_y": -0.5},
            {"theta_E": 0.02, "center_x": -0.4, "center_y": -0.4},
        ]

        self.kwargs_epl = kwargs_macro + halo_kwargs
        self.zlist_epl = [self.zlens, self.zlens] + halo_z
        self.lens_model_list_epl = ["EPL", "SHEAR"] + halo_list

        self.lensModel = LensModel(
            self.lens_model_list_epl,
            self.zlens,
            self.zsource,
            self.zlist_epl,
            multi_plane=True,
        )

        self.param_class = PowerLawFreeShear(self.kwargs_epl)

    def test_rayshooting(self):
        solver = LensEquationSolver(self.lensModel)
        source_x, source_y = -0.05, -0.02
        x_image_true, y_image_true = solver.findBrightImage(
            source_x, source_y, self.kwargs_epl
        )

        fast_rayshooting = MultiplaneFast(
            x_image_true,
            y_image_true,
            self.zlens,
            self.zsource,
            self.lens_model_list_epl,
            self.zlist_epl,
            astropy_instance=None,
            param_class=self.param_class,
            foreground_rays=None,
            tol_source=1e-5,
            numerical_alpha_class=None,
        )

        (
            x_fore,
            y_fore,
            alpha_x_fore,
            alpha_y_fore,
        ) = fast_rayshooting._ray_shooting_fast_foreground()
        (
            xtrue,
            ytrue,
            alpha_xtrue,
            alpha_ytrue,
        ) = self.lensModel.lens_model.ray_shooting_partial(
            np.zeros_like(x_image_true),
            np.zeros_like(y_image_true),
            x_image_true,
            y_image_true,
            0.0,
            self.zlens,
            self.kwargs_epl,
        )

        npt.assert_almost_equal(x_fore, xtrue)
        npt.assert_almost_equal(y_fore, ytrue)

        args_lens = self.param_class.kwargs_to_args(self.kwargs_epl)
        xfast, yfast = fast_rayshooting.ray_shooting_fast(args_lens)

        x, y = self.lensModel.ray_shooting(x_image_true, y_image_true, self.kwargs_epl)

        npt.assert_almost_equal(xfast, x)
        npt.assert_almost_equal(yfast, y)

        x_inner, y_inner = fast_rayshooting.lensModel.ray_shooting(
            x_image_true, y_image_true, self.kwargs_epl
        )
        npt.assert_almost_equal(x_inner, xfast)
        npt.assert_almost_equal(y_inner, yfast)

        foreground_rays = fast_rayshooting._foreground_rays

        fast_rayshooting_new = MultiplaneFast(
            x_image_true,
            y_image_true,
            self.zlens,
            self.zsource,
            self.lens_model_list_epl,
            self.zlist_epl,
            astropy_instance=None,
            param_class=self.param_class,
            foreground_rays=foreground_rays,
            tol_source=1e-5,
            numerical_alpha_class=None,
        )

        xfast, yfast = fast_rayshooting_new.ray_shooting_fast(args_lens)

        npt.assert_almost_equal(xfast, x)
        npt.assert_almost_equal(yfast, y)

        chi_square_source = fast_rayshooting.source_plane_chi_square(args_lens)
        chi_square_total = fast_rayshooting.chi_square(args_lens)

        logL = fast_rayshooting.logL(args_lens)
        logL_true = -0.5 * chi_square_total
        npt.assert_almost_equal(logL, logL_true)
        npt.assert_almost_equal(chi_square_total, chi_square_source)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import unittest
import pytest

from lenstronomy.Sampling.parameters import Param
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver


class TestParam(object):
    def setup_method(self):
        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "source_light_model_list": ["GAUSSIAN"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "multi_plane": True,
            "lens_redshift_list": [0.5],
            "z_source": 2,
            "source_redshift_list": [0.5],
        }
        kwargs_param = {
            "num_point_source_list": [2],
            "lens_redshift_sampling_indexes": [0],
            "source_redshift_sampling_indexes": [0],
            "image_plane_source_list": [True],
        }
        kwargs_fixed_lens = [{"gamma": 1.9}]  # for SPEP lens
        kwargs_fixed_source = [{"sigma": 0.1, "center_x": 0.2, "center_y": 0.2}]
        kwargs_fixed_ps = [{"ra_image": [-1, 1], "dec_image": [-1, 1]}]
        kwargs_fixed_lens_light = [{}]
        kwargs_fixed_cosmo = [{}]
        self.param_class = Param(
            kwargs_model,
            kwargs_fixed_lens=kwargs_fixed_lens,
            kwargs_fixed_source=kwargs_fixed_source,
            kwargs_fixed_lens_light=kwargs_fixed_lens_light,
            kwargs_fixed_ps=kwargs_fixed_ps,
            kwargs_fixed_special=kwargs_fixed_cosmo,
            **kwargs_param
        )
        self.param_class.print_setting()

    def test_num_param(self):
        num_param, list = self.param_class.num_param()
        assert list[0] == "theta_E_lens0"
        assert num_param == 10

        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "source_light_model_list": ["GAUSSIAN"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        kwargs_param = {}
        kwargs_fixed_lens = [{"gamma": 1.9}]  # for SPEP lens
        kwargs_fixed_source = [{"sigma": 0.1, "center_x": 0.2, "center_y": 0.2}]
        kwargs_fixed_ps = [{"ra_image": [-1, 1], "dec_image": [-1, 1]}]
        kwargs_fixed_lens_light = [{}]
        kwargs_fixed_cosmo = [{}]
        param_class_linear = Param(
            kwargs_model,
            kwargs_fixed_lens,
            kwargs_fixed_source,
            kwargs_fixed_lens_light,
            kwargs_fixed_ps,
            kwargs_fixed_cosmo,
            linear_solver=True,
            **kwargs_param
        )
        num_param, list = param_class_linear.num_param()
        assert list[0] == "theta_E_lens0"
        print(list)
        assert len(list) == num_param
        assert num_param == 9

    def test_num_param_linear(self):
        num_param = self.param_class.num_param_linear()
        assert num_param == 4

    def test_get_params(self):
        kwargs_true_lens = [
            {
                "theta_E": 1.0,
                "gamma": 1.9,
                "e1": 0.01,
                "e2": -0.01,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]  # for SPEP lens
        kwargs_true_source = [
            {
                "amp": 1 * 2 * np.pi * 0.1**2,
                "center_x": 0.2,
                "center_y": 0.2,
                "sigma": 0.1,
            }
        ]
        kwargs_true_lens_light = [
            {
                "center_x": -0.06,
                "center_y": 0.4,
                "phi_G": 4.8,
                "q": 0.86,
                "n_sersic": 1.7,
                "amp": 11.8,
                "R_sersic": 0.697,
                "phi_G_2": 0,
            }
        ]
        kwargs_true_ps = [
            {"point_amp": [1, 1], "ra_image": [-1, 1], "dec_image": [-1, 1]}
        ]
        kwargs_special = {"z_sampling": [0.5]}
        args = self.param_class.kwargs2args(
            kwargs_true_lens,
            kwargs_true_source,
            kwargs_lens_light=kwargs_true_lens_light,
            kwargs_ps=kwargs_true_ps,
            kwargs_special=kwargs_special,
        )
        kwargs_return = self.param_class.args2kwargs(args)
        lens_dict_list = kwargs_return["kwargs_lens"]
        lens_light_dict_list = kwargs_return["kwargs_lens_light"]
        lens_dict = lens_dict_list[0]
        assert lens_dict["theta_E"] == 1.0
        assert lens_dict["gamma"] == 1.9
        assert lens_dict["e1"] == 0.01
        assert lens_dict["e2"] == -0.01
        assert lens_dict["center_x"] == 0.0
        assert lens_dict["center_y"] == 0.0
        assert lens_light_dict_list[0]["center_x"] == -0.06

    def test_get_cosmo(self):
        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "source_light_model_list": ["GAUSSIAN"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        kwargs_param = {"Ddt_sampling": True}
        kwargs_fixed_lens = [{"gamma": 1.9}]  # for SPEP lens
        kwargs_fixed_source = [{"sigma": 0.1, "center_x": 0.2, "center_y": 0.2}]
        kwargs_fixed_ps = [{"ra_image": [-1, 1], "dec_image": [-1, 1]}]
        kwargs_fixed_lens_light = [{}]
        kwargs_fixed_cosmo = {"D_dt": 1000}
        param_class = Param(
            kwargs_model,
            kwargs_fixed_lens=kwargs_fixed_lens,
            kwargs_fixed_source=kwargs_fixed_source,
            kwargs_fixed_lens_light=kwargs_fixed_lens_light,
            kwargs_fixed_ps=kwargs_fixed_ps,
            kwargs_fixed_special=kwargs_fixed_cosmo,
            **kwargs_param
        )

        kwargs_true_lens = [
            {
                "theta_E": 1.0,
                "gamma": 1.9,
                "e1": 0.01,
                "e2": -0.01,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]  # for SPEP lens
        kwargs_true_source = [
            {
                "amp": 1 * 2 * np.pi * 0.1**2,
                "center_x": 0.2,
                "center_y": 0.2,
                "sigma": 0.1,
            }
        ]
        kwargs_true_lens_light = [
            {
                "center_x": -0.06,
                "center_y": 0.4,
                "phi_G": 4.8,
                "q": 0.86,
                "n_sersic": 1.7,
                "amp": 11.8,
                "R_sersic": 0.697,
                "phi_G_2": 0,
            }
        ]
        kwargs_true_ps = [
            {"point_amp": [1, 1], "ra_image": [-1, 1], "dec_image": [-1, 1]}
        ]
        args = param_class.kwargs2args(
            kwargs_true_lens,
            kwargs_true_source,
            kwargs_lens_light=kwargs_true_lens_light,
            kwargs_ps=kwargs_true_ps,
            kwargs_special={"D_dt": 1000},
        )
        assert param_class.specialParams._D_dt_sampling.on

    def test_mass_scaling(self):
        kwargs_model = {
            "lens_model_list": ["SIS", "NFW", "NFW", "SIS", "SERSIC", "HERNQUIST"]
        }
        kwargs_constraints = {"mass_scaling_list": [False, 1, 1, 1, 1, 1]}
        kwargs_fixed_lens = [
            {},
            {"alpha_Rs": 0.1},
            {"alpha_Rs": 0.3},
            {"theta_E": 0.1},
            {"k_eff": 0.3},
            {"sigma0": 1},
        ]
        kwargs_fixed_cosmo = {}
        param_class = Param(
            kwargs_model,
            kwargs_fixed_lens=kwargs_fixed_lens,
            kwargs_fixed_special=kwargs_fixed_cosmo,
            **kwargs_constraints
        )
        kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"alpha_Rs": 0.1, "Rs": 5, "center_x": 1.0, "center_y": 0},
            {"alpha_Rs": 0.1, "Rs": 5, "center_x": 0, "center_y": 1.0},
            {"theta_E": 0.1, "center_x": 3, "center_y": 1.0},
            {
                "k_eff": 0.3,
                "R_sersic": 1,
                "n_sersic": 2,
                "center_x": 3,
                "center_y": 1.0,
            },
            {"sigma0": 1, "Rs": 1, "center_x": 3, "center_y": 1.0},
        ]
        kwargs_source = []
        kwargs_lens_light = []
        kwargs_ps = []
        mass_scale = 2
        kwargs_cosmo = {"scale_factor": [mass_scale]}
        args = param_class.kwargs2args(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_special=kwargs_cosmo,
        )
        assert args[-1] == mass_scale

        kwargs_return = param_class.args2kwargs(args)
        kwargs_lens = kwargs_return["kwargs_lens"]
        print(kwargs_lens, "test")
        assert kwargs_lens[0]["theta_E"] == 1
        assert kwargs_lens[1]["alpha_Rs"] == 0.1 * mass_scale
        assert kwargs_lens[2]["alpha_Rs"] == 0.3 * mass_scale
        assert kwargs_lens[3]["theta_E"] == 0.1 * mass_scale
        assert kwargs_lens[4]["k_eff"] == 0.3 * mass_scale
        assert kwargs_lens[5]["sigma0"] == 1 * mass_scale

        kwargs_return = param_class.args2kwargs(args, bijective=True)
        kwargs_lens = kwargs_return["kwargs_lens"]
        assert kwargs_lens[0]["theta_E"] == 1
        assert kwargs_lens[1]["alpha_Rs"] == 0.1
        assert kwargs_lens[2]["alpha_Rs"] == 0.3

    def test_general_scaling(self):
        kwargs_model = {"lens_model_list": ["PJAFFE", "PJAFFE", "NFW", "PJAFFE", "NFW"]}
        # Scale Rs for two of the PJAFFEs, and sigma0 for a different set of PJAFFEs
        # Scale alpha_Rs for the NFWs
        kwargs_constraints = {
            "general_scaling": {
                "Rs": [1, False, False, 1, False],
                "sigma0": [False, 1, False, 1, False],
                "alpha_Rs": [False, False, 1, False, 1],
            }
        }
        # PJAFFE: sigma0, Ra, Rs, center_x, center_y
        # NFW: Rs, alpha_Rs, center_x, center_y
        kwargs_fixed_lens = [
            {"Rs": 2.0, "center_x": 1.0},
            {"sigma0": 2.0, "Ra": 2.0, "Rs": 3.0, "center_y": 1.5},
            {"alpha_Rs": 0.1},
            {"Ra": 0.1, "center_x": 0, "center_y": 0},
            {"Rs": 3, "center_x": -1, "center_y": 3},
        ]
        kwargs_fixed_cosmo = {}
        param_class = Param(
            kwargs_model,
            kwargs_fixed_lens=kwargs_fixed_lens,
            kwargs_fixed_special=kwargs_fixed_cosmo,
            **kwargs_constraints
        )
        kwargs_lens = [
            {"sigma0": 3, "Ra": 2, "center_y": 5},
            {"center_x": 1.0},
            {"Rs": 3, "center_x": 0.0, "center_y": -1.0},
            {"sigma0": 3, "Rs": 1.5},
            {"alpha_Rs": 4},
        ]
        kwargs_source = []
        kwargs_lens_light = []
        kwargs_ps = []
        # Define the scaling and power for each parameter
        kwargs_cosmo = {
            "Rs_scale_factor": [2.0],
            "Rs_scale_pow": [1.1],
            "sigma0_scale_factor": [3],
            "sigma0_scale_pow": [2.0],
            "alpha_Rs_scale_factor": [0.3],
            "alpha_Rs_scale_pow": [0.5],
        }
        args = param_class.kwargs2args(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_special=kwargs_cosmo,
        )
        num, names = param_class.num_param()
        print(names)
        print(args)

        kwargs_return = param_class.args2kwargs(args)
        kwargs_lens = kwargs_return["kwargs_lens"]
        print("kwargs_lens:", kwargs_lens)
        npt.assert_almost_equal(kwargs_lens[0]["Rs"], 2.0 * 2.0**1.1)
        npt.assert_almost_equal(kwargs_lens[0]["sigma0"], 3)
        npt.assert_almost_equal(kwargs_lens[1]["Rs"], 3.0)
        npt.assert_almost_equal(kwargs_lens[1]["sigma0"], 3.0 * 2.0**2.0)
        npt.assert_almost_equal(kwargs_lens[2]["alpha_Rs"], 0.3 * 0.1**0.5)
        npt.assert_almost_equal(kwargs_lens[3]["Rs"], 2.0 * 1.5**1.1)
        npt.assert_almost_equal(kwargs_lens[3]["sigma0"], 3.0 * 3**2.0)
        npt.assert_almost_equal(kwargs_lens[4]["alpha_Rs"], 0.3 * 4**0.5)

        kwargs_return = param_class.args2kwargs(args, bijective=True)
        kwargs_lens = kwargs_return["kwargs_lens"]
        npt.assert_almost_equal(kwargs_lens[0]["Rs"], 2.0)
        npt.assert_almost_equal(kwargs_lens[1]["sigma0"], 2.0)
        npt.assert_almost_equal(kwargs_lens[2]["alpha_Rs"], 0.1)
        npt.assert_almost_equal(kwargs_lens[3]["Rs"], 1.5)
        npt.assert_almost_equal(kwargs_lens[3]["sigma0"], 3)
        npt.assert_almost_equal(kwargs_lens[4]["alpha_Rs"], 4)

    def test_joint_lens_with_light(self):
        kwargs_model = {
            "lens_model_list": ["CHAMELEON"],
            "lens_light_model_list": ["CHAMELEON"],
        }
        i_light, k_lens = 0, 0
        kwargs_constraints = {
            "joint_lens_with_light": [
                [i_light, k_lens, ["w_t", "w_c", "center_x", "center_y", "e1", "e2"]]
            ]
        }
        kwargs_lens = [{"alpha_1": 10}]
        kwargs_lens_light = [
            {
                "amp": 1,
                "w_t": 0.5,
                "w_c": 0.1,
                "center_x": 0,
                "center_y": 0.3,
                "e1": 0.1,
                "e2": -0.2,
            }
        ]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens, kwargs_lens_light=kwargs_lens_light
        )
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        kwargs_lens_light_out = kwargs_return["kwargs_lens_light"]
        assert kwargs_lens_out[0]["w_c"] == kwargs_lens_light[0]["w_c"]
        assert kwargs_lens_light_out[0]["w_c"] == kwargs_lens_light[0]["w_c"]

        kwargs_model = {"lens_model_list": ["SIS"], "lens_light_model_list": ["SERSIC"]}
        i_light, k_lens = 0, 0
        kwargs_constraints = {
            "joint_lens_with_light": [[i_light, k_lens, ["center_x", "center_y"]]]
        }  # list[[i_point_source, k_source, ['param_name1', 'param_name2', ...]], [
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [
            {"amp": 1, "R_sersic": 0.5, "n_sersic": 2, "center_x": 1, "center_y": 1}
        ]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens, kwargs_lens_light=kwargs_lens_light
        )
        # kwargs_lens_out, kwargs_source_out, _, kwargs_ps_out, _ = param.args2kwargs(args)
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]

        assert kwargs_lens_out[0]["theta_E"] == kwargs_lens[0]["theta_E"]
        assert kwargs_lens_out[0]["center_x"] == kwargs_lens_light[0]["center_x"]

    def test_joint_lens_with_light_dict(self):
        kwargs_model = {
            "lens_model_list": ["SHEAR"],
            "lens_light_model_list": ["SERSIC"],
        }
        i_light, k_lens = 0, 0
        kwargs_constraints = {
            "joint_lens_with_light": [
                [i_light, k_lens, {"ra_0": "center_x", "dec_0": "center_y"}]
            ]
        }
        kwargs_lens = [{"gamma1": 0.05, "gamma2": 0.06}]
        kwargs_lens_light = [
            {"amp": 1, "R_sersic": 1, "n_sersic": 4, "center_x": 0.1, "center_y": 0.3}
        ]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens, kwargs_lens_light=kwargs_lens_light
        )
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        kwargs_lens_light_out = kwargs_return["kwargs_lens_light"]
        assert kwargs_lens_out[0]["gamma1"] == kwargs_lens[0]["gamma1"]
        assert kwargs_lens_out[0]["ra_0"] == kwargs_lens_light[0]["center_x"]
        assert kwargs_lens_out[0]["dec_0"] == kwargs_lens_light[0]["center_y"]

    def test_joint_source_with_point_source(self):
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["SOURCE_POSITION"],
        }
        i_source, k_ps = 0, 0
        kwargs_constraints = {
            "joint_source_with_point_source": [[k_ps, i_source]]
        }  # list[[i_point_source, k_source, ['param_name1', 'param_name2', ...]], [

        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_source = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 1, "center_y": 1}
        ]
        kwargs_ps = [{"ra_source": 0.5, "dec_source": 0.5}]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens, kwargs_source=kwargs_source, kwargs_ps=kwargs_ps
        )
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        kwargs_source_out = kwargs_return["kwargs_source"]
        # kwargs_lens_out, kwargs_source_out, _, kwargs_ps_out, _ = param.args2kwargs(args)
        assert kwargs_lens_out[0]["theta_E"] == kwargs_lens[0]["theta_E"]
        assert kwargs_source_out[0]["center_x"] == kwargs_ps[0]["ra_source"]

        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        i_source, k_ps = 0, 0
        kwargs_constraints = {
            "joint_source_with_point_source": [[k_ps, i_source]]
        }  # list[[i_point_source, k_source, ['param_name1', 'param_name2', ...]], [

        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_source = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 1, "center_y": 1}
        ]
        kwargs_ps = [{"ra_image": [0.5], "dec_image": [0.5]}]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens, kwargs_source=kwargs_source, kwargs_ps=kwargs_ps
        )
        # kwargs_lens_out, kwargs_source_out, _, kwargs_ps_out, _ = param.args2kwargs(args)
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        kwargs_source_out = kwargs_return["kwargs_source"]
        assert kwargs_lens_out[0]["theta_E"] == kwargs_lens[0]["theta_E"]
        npt.assert_almost_equal(kwargs_source_out[0]["center_x"], -0.207, decimal=2)

    def test_joint_lens_light_with_point_source(self):
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "lens_light_model_list": ["SERSIC"],
        }
        i_lens_light, k_ps = 0, 0
        kwargs_constraints = {
            "joint_lens_light_with_point_source": [[k_ps, i_lens_light]]
        }  # list[[i_point_source, k_source, ['param_name1', 'param_name2', ...]], [

        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_source = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 1, "center_y": 1}
        ]
        kwargs_lens_light = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 0.2, "center_y": 0.2}
        ]
        kwargs_ps = [{"ra_image": [0.5], "dec_image": [0.5]}]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
        )
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_light_out = kwargs_return["kwargs_lens_light"]
        kwargs_ps_out = kwargs_return["kwargs_ps"]
        assert kwargs_lens_light_out[0]["center_x"] == kwargs_ps_out[0]["ra_image"]

    def test_logsampling(self):
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "lens_light_model_list": ["SERSIC"],
        }
        kwargs_constraints = {"log_sampling_lens": [[0, ["theta_E"]]]}

        kwargs_lens = [{"theta_E": 0.1, "center_x": 0, "center_y": 0}]
        kwargs_source = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 1, "center_y": 1}
        ]
        kwargs_lens_light = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 0.2, "center_y": 0.2}
        ]
        kwargs_ps = [{"ra_image": [0.5], "dec_image": [0.5]}]
        param = Param(kwargs_model=kwargs_model, **kwargs_constraints)
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
        )
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        assert args[0] == -1
        assert kwargs_lens_out[0]["theta_E"] == 0.1

    def test_with_solver(self):
        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        i_lens_light, k_ps = 0, 0
        kwargs_constraints = {"solver_type": "PROFILE", "num_point_source_list": [4]}

        kwargs_lens = [
            {
                "theta_E": 1,
                "gamma": 2,
                "e1": 0.1,
                "e2": 0.1,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        kwargs_source = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 1, "center_y": 1}
        ]
        kwargs_lens_light = [
            {"amp": 1, "n_sersic": 2, "R_sersic": 0.3, "center_x": 0.2, "center_y": 0.2}
        ]
        lensModel = LensModel(lens_model_list=["SPEP"])
        lensEquationSlover = LensEquationSolver(lensModel=lensModel)
        x_image, y_image = lensEquationSlover.image_position_from_source(
            sourcePos_x=0.0, sourcePos_y=0.01, kwargs_lens=kwargs_lens
        )
        print(x_image, y_image, "test")
        kwargs_ps = [{"ra_image": x_image, "dec_image": y_image}]
        param = Param(
            kwargs_model=kwargs_model,
            kwargs_lens_init=kwargs_lens,
            **kwargs_constraints
        )
        args = param.kwargs2args(
            kwargs_lens=kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
        )
        # kwargs_lens_out, kwargs_source_out, kwargs_lens_light_out, kwargs_ps_out, _ = param.args2kwargs(args)
        kwargs_return = param.args2kwargs(args)
        kwargs_lens_out = kwargs_return["kwargs_lens"]
        kwargs_ps_out = kwargs_return["kwargs_ps"]
        dist = param.check_solver(kwargs_lens=kwargs_lens_out, kwargs_ps=kwargs_ps_out)
        npt.assert_almost_equal(dist, 0, decimal=10)

    def test_num_point_source_images(self):
        num = self.param_class.num_point_source_images
        assert num == 2

    def test_shapelet_lens(self):
        kwargs_model = {
            "lens_model_list": ["SHAPELETS_CART"],
            "source_light_model_list": [],
            "lens_light_model_list": [],
            "point_source_model_list": [],
        }
        kwargs_param = {"num_shapelet_lens": 6}
        kwargs_fixed_lens = [{"beta": 1}]  # for SPEP lens
        kwargs_fixed_source = [{}]
        kwargs_fixed_ps = [{}]
        kwargs_fixed_lens_light = [{}]
        kwargs_fixed_cosmo = [{}]
        self.param_class = Param(
            kwargs_model,
            kwargs_fixed_lens=kwargs_fixed_lens,
            kwargs_fixed_source=kwargs_fixed_source,
            kwargs_fixed_lens_light=kwargs_fixed_lens_light,
            kwargs_fixed_ps=kwargs_fixed_ps,
            kwargs_fixed_special=kwargs_fixed_cosmo,
            **kwargs_param
        )
        self.param_class.print_setting()


class TestRaise(unittest.TestCase):
    def test_raise(self):
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "lens_light_model_list": ["SERSIC"],
        }
        kwargs_constraints = {
            "log_sampling_lens": [[0, {"theta_E"}]]
        }  # wrong type, dict instead of list
        with self.assertRaises(TypeError):
            param = Param(kwargs_model=kwargs_model, **kwargs_constraints)


if __name__ == "__main__":
    pytest.main()



import numpy as np

from lenstronomy.Data.pixel_grid import PixelGrid
from lenstronomy.Data.image_noise import ImageNoise

__all__ = ["ImageData"]


class ImageData(PixelGrid, ImageNoise):
    """Class to handle the data, coordinate system and masking, including convolution
    with various numerical precisions.

    The Data() class is initialized with keyword arguments:

    - 'image_data': 2d numpy array of the image data
    - 'transform_pix2angle' 2x2 transformation matrix (linear) to transform a pixel shift into a coordinate shift (x, y) -> (ra, dec)
    - 'ra_at_xy_0' RA coordinate of pixel (0,0)
    - 'dec_at_xy_0' DEC coordinate of pixel (0,0)

    optional keywords for shifts in the coordinate system:

    - 'ra_shift': shifts the coordinate system with respect to 'ra_at_xy_0'
    - 'dec_shift': shifts the coordinate system with respect to 'dec_at_xy_0'

    optional keywords for noise properties:

    - 'background_rms': rms value of the background noise
    - 'exp_time': float, exposure time to compute the Poisson noise contribution
    - 'exposure_map': 2d numpy array, effective exposure time for each pixel. If set, will replace 'exp_time'
    - 'noise_map': Gaussian noise (1-sigma) for each individual pixel.

    If this keyword is set, the other noise properties will be ignored.

    optional keywords for interferometric quantities:

    - 'likelihood_method': need to be specified to 'interferometry_natwt' if one needs to use the interferometric likelihood function.

    The default of 'likelihood_method' is 'diagonal', which is used for non-correlated noises (usually for the CCD images.)

    - 'log_likelihood_constant': a constant that adds to logL.
    - 'antenna_primary_beam': primary beam pattern of antennae (now treat each antenna dish with the same primary beam).

    ** notes **
    the likelihood for the data given model P(data|model) is defined in the function below. Please make sure that
    your definitions and units of 'exposure_map', 'background_rms' and 'image_data' are in accordance with the
    likelihood function. In particular, make sure that the Poisson noise contribution is defined in the count rate.
    """

    def __init__(
        self,
        image_data,
        exposure_time=None,
        background_rms=None,
        noise_map=None,
        gradient_boost_factor=None,
        ra_at_xy_0=0,
        dec_at_xy_0=0,
        transform_pix2angle=None,
        ra_shift=0,
        dec_shift=0,
        phi_rot=0,
        log_likelihood_constant=0,
        antenna_primary_beam=None,
        likelihood_method="diagonal",
        flux_scaling=1,
    ):
        """

        :param image_data: 2d numpy array of the image data
        :param exposure_time: int or array of size the data; exposure time
         (common for all pixels or individually for each individual pixel)
        :param background_rms: root-mean-square value of Gaussian background noise in units counts per second
        :param noise_map: int or array of size the data; joint noise sqrt(variance) of each individual pixel.
        :param gradient_boost_factor: None or float, variance terms added in quadrature scaling with
         gradient^2 * gradient_boost_factor
        :param transform_pix2angle: 2x2 matrix, mapping of pixel to coordinate
        :param ra_at_xy_0: ra coordinate at pixel (0,0)
        :param dec_at_xy_0: dec coordinate at pixel (0,0)
        :param ra_shift: RA shift of pixel grid
        :param dec_shift: DEC shift of pixel grid
        :param log_likelihood_constant: float, allows user to input a constant that will be added to the log likelihood. Note that, as for now, this variable is ONLY used for interferometric mode.
        :param antenna_primary_beam: 2d numpy array with the same size of imaga_data;
        :param phi_rot: rotation angle in regard to pixel coordinate transform_pix2angle
        :param antenna_primary_beam: 2d numpy array with the same size of image_data;
         more descriptions of the primary beam can be found in the AngularSensitivity class
        :param likelihood_method: string, type of method of log_likelihood computation: options are 'diagonal', 'interferometry_natwt'.
         The default option 'diagonal' uses a diagonal covariance matrix, which is the case for CCD images.
         The 'interferometry_natwt' option uses our special interferometric likelihood function based on natural weighting images.
        :param flux_scaling: scales the model amplitudes to match the imaging data units. This can be used, for example,
         when modeling multiple exposures that have different magnitude zero points (or flux normalizations) but demand
         the same model normalization
        """
        nx, ny = np.shape(image_data)
        if transform_pix2angle is None:
            transform_pix2angle = np.array([[1, 0], [0, 1]])
        cos_phi, sin_phi = np.cos(phi_rot), np.sin(phi_rot)
        rot_matrix = np.array([[cos_phi, -sin_phi], [sin_phi, cos_phi]])
        transform_pix2angle_rot = np.dot(transform_pix2angle, rot_matrix)
        PixelGrid.__init__(
            self,
            nx,
            ny,
            transform_pix2angle_rot,
            ra_at_xy_0 + ra_shift,
            dec_at_xy_0 + dec_shift,
            antenna_primary_beam,
        )
        ImageNoise.__init__(
            self,
            image_data,
            exposure_time=exposure_time,
            background_rms=background_rms,
            noise_map=noise_map,
            gradient_boost_factor=gradient_boost_factor,
            verbose=False,
            flux_scaling=flux_scaling,
        )

        self._logL_constant = log_likelihood_constant
        self._logL_method = likelihood_method
        if (
            self._logL_method != "diagonal"
            and self._logL_method != "interferometry_natwt"
        ):
            raise ValueError(
                "likelihood_method %s not supported! likelihood_method can only be 'diagonal' or 'interferometry_natwt'!"
                % self._logL_method
            )

    def update_data(self, image_data):
        """Update the data as well as the error matrix estimated from it when done so
        using the data.

        :param image_data: 2d numpy array of same size as nx, ny
        :return: None
        """
        nx, ny = np.shape(image_data)
        if not self._nx == nx and not self._ny == ny:
            raise ValueError(
                "shape of new data %s %s must equal old data %s %s!"
                % (nx, ny, self._nx, self._ny)
            )
        self._data = image_data
        if hasattr(self, "_C_D") and self._noise_map is None:
            del self._C_D

    @property
    def data(self):
        """

        :return: 2d numpy array of data
        """
        return self._data

    def log_likelihood(self, model, mask, additional_error_map=0):
        """Computes the likelihood of the data given the model p(data|model) The
        Gaussian errors are estimated with the covariance matrix, based on the model
        image. The errors include the background rms value and the exposure time to
        compute the Poisson noise level (in Gaussian approximation).

        :param model: the model (same dimensions and units as data)
        :param mask: bool (1, 0) values per pixel. If =0, the pixel is ignored in the
            likelihood
        :param additional_error_map: additional error term (in same units as covariance
            matrix). This can e.g. come from model errors in the PSF estimation.
        :return: the natural logarithm of the likelihood p(data|model)
        """
        # if the likelihood method is assigned to be 'interferometry_natwt', it will return logL computed using the interfermetric likelihood function
        if self._logL_method == "interferometry_natwt":
            return self.log_likelihood_interferometry(model)

        c_d = self.C_D_model(model)
        chi2 = (model - self._data) ** 2 / (c_d + np.abs(additional_error_map)) * mask
        chi2 = np.array(chi2)
        log_likelihood = -np.sum(chi2) / 2
        return log_likelihood

    def log_likelihood_interferometry(self, model):
        """log_likelihood function for natural weighting interferometric images, based
        on (placeholder for Nan Zhang's paper).

        For the interferometry case, the model should be in the form [array1, array2],
        where array1 and array2 are unconvolved and convolved model images respectively.
        They are both 2d array with the same shape of the data.

        The chi^2 of interferometry is computed by

        .. math::
            \\chi^2 =  (d-Ax)^TC^{-1}(d-Ax) = \\frac{1}{\\sigma^2}(d^TA^{-1}d - 2x^Td + x^TAx)

        where :math:`d` and :math:`x` are the data vector and the unconvolved model image vector respectively.
        :math:`A` is the convolution operation matrix, where we normalize the PSF by setting its central pixel to 1.
        :math:`C` is the noise covariance matrix, its diagonal entries are rms^2 of noises, :math:`\\sigma^2`.
        For natural weighting interferometric images, we used the relation
        (see Section 3.2 of https://doi.org/10.1093/mnras/staa2740 for the relation of natural weighting covariance matrix and PSF convolution)

        .. math::
            C = \\sigma^2 A

        to simplify the likelihood function above.
        """

        xd = np.sum(model[0] * self._data)
        xAx = np.sum(model[0] * model[1])
        logL = -(xAx - 2 * xd) / (2 * self._background_rms**2) + self._logL_constant
        return logL

    def likelihood_method(self):
        """Pass the likelihood_method to the ImageModel and will be used to identify the
        method of likelihood computation in ImageLinearFit.

        :return: string, likelihood method
        """
        return self._logL_method



import pytest
import copy
import numpy.testing as npt
from lenstronomy.LensModel.Profiles.curved_arc_spt import CurvedArcSPT
from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import CurvedArcSISMST
from lenstronomy.Util import util


class TestCurvedArcSPT(object):
    def setup_method(self):
        self._curve_spt = CurvedArcSPT()
        self._curve_regular = CurvedArcSISMST()

    def test_function(self):
        kwargs_arc = {
            "tangential_stretch": 5,
            "radial_stretch": 1,
            "curvature": 1.0 / 10,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
            "gamma1": 0,
            "gamma2": 0,
        }
        npt.assert_raises(Exception, self._curve_spt.function, 0.0, 0.0, **kwargs_arc)

    def test_spt_mapping(self):
        e1, e2 = 0.1, -0.2

        kwargs_arc_sis_mst = {
            "tangential_stretch": 3,
            "radial_stretch": 1.2,
            "curvature": 0.8,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }

        # inverse reduced shear transform as SPT
        kwargs_arc_spt = copy.deepcopy(kwargs_arc_sis_mst)
        kwargs_arc_spt["gamma1"] = -e1
        kwargs_arc_spt["gamma2"] = -e2

        x, y = util.make_grid(numPix=100, deltapix=0.01)
        f_x_sis, f_y_sis = self._curve_regular.derivatives(x, y, **kwargs_arc_sis_mst)
        beta_x_sis = x - f_x_sis
        beta_y_sis = y - f_y_sis
        f_x_spt, f_y_spt = self._curve_spt.derivatives(x, y, **kwargs_arc_spt)
        beta_x_spt = x - f_x_spt
        beta_y_spt = y - f_y_spt

        from lenstronomy.LightModel.light_model import LightModel

        gauss = LightModel(["GAUSSIAN_ELLIPSE"])
        kwargs_source = [
            {"amp": 1, "sigma": 0.05, "center_x": 0, "center_y": 0, "e1": 0, "e2": 0}
        ]
        kwargs_source_spt = copy.deepcopy(kwargs_source)
        kwargs_source_spt[0]["e1"] = e1
        kwargs_source_spt[0]["e2"] = e2
        flux_sis = gauss.surface_brightness(beta_x_sis, beta_y_sis, kwargs_source)
        flux_spt = gauss.surface_brightness(beta_x_spt, beta_y_spt, kwargs_source_spt)
        npt.assert_almost_equal(flux_sis, flux_spt)


if __name__ == "__main__":
    pytest.main("-k TestCurvedArcSPT")



import numpy as np
import lenstronomy.Util.util as util
import lenstronomy.Util.image_util as image_util
from scipy.optimize import minimize
from lenstronomy.LensModel.Solver.epl_shear_solver import solve_lenseq_pemd

__all__ = ["LensEquationSolver"]


class LensEquationSolver(object):
    """Class to solve for image positions given lens model and source position."""

    def __init__(self, lensModel):
        """This class must contain the following definitions (with same syntax as the
        standard LensModel() class: def ray_shooting() def hessian() def magnification()

        :param lensModel: instance of a class according to
            lenstronomy.LensModel.lens_model
        """
        self.lensModel = lensModel

    def image_position_stochastic(
        self,
        source_x,
        source_y,
        kwargs_lens,
        search_window=10,
        precision_limit=10 ** (-10),
        arrival_time_sort=True,
        x_center=0,
        y_center=0,
        num_random=1000,
    ):
        """Solves the lens equation stochastic with the scipy minimization routine on
        the quadratic distance between the backwards ray-shooted proposed image position
        and the source position. Credits to Giulia Pagano.

        :param source_x: source position
        :param source_y: source position
        :param kwargs_lens: lens model list of keyword arguments
        :param search_window: angular size of search window
        :param precision_limit: limit required on the precision in the source plane
        :param arrival_time_sort: bool, if True sorts according to arrival time
        :param x_center: center of search window
        :param y_center: center of search window
        :param num_random: number of random starting points of the non-linear solver in
            the search window
        :return: x_image, y_image
        """
        kwargs_lens = self.lensModel.set_static(kwargs_lens)

        x_solve, y_solve = [], []
        for i in range(num_random):
            x_init = (
                np.random.uniform(-search_window / 2.0, search_window / 2) + x_center
            )
            y_init = (
                np.random.uniform(-search_window / 2.0, search_window / 2) + y_center
            )
            xinitial = np.array([x_init, y_init])
            result = minimize(
                self._root,
                xinitial,
                args=(kwargs_lens, source_x, source_y),
                tol=precision_limit**2,
                method="Nelder-Mead",
            )
            if (
                self._root(result.x, kwargs_lens, source_x, source_y)
                < precision_limit**2
            ):
                x_solve.append(result.x[0])
                y_solve.append(result.x[1])

        x_mins, y_mins = image_util.findOverlap(x_solve, y_solve, precision_limit)
        if arrival_time_sort:
            x_mins, y_mins = self.sort_arrival_times(x_mins, y_mins, kwargs_lens)
        self.lensModel.set_dynamic()
        return x_mins, y_mins

    def _root(self, x, kwargs_lens, source_x, source_y):
        """

        :param x: parameters [x-coord, y-coord]
        :param kwargs_lens: list of keyword arguments of the lens model
        :param source_x: source position
        :param source_y: source position
        :return: square distance between ray-traced image position and given source position
        """
        x_, y_ = x
        beta_x, beta_y = self.lensModel.ray_shooting(x_, y_, kwargs_lens)
        return (beta_x - source_x) ** 2 + (beta_y - source_y) ** 2

    def candidate_solutions(
        self,
        sourcePos_x,
        sourcePos_y,
        kwargs_lens,
        min_distance=0.1,
        search_window=10,
        verbose=False,
        x_center=0,
        y_center=0,
    ):
        """Finds pixels in the image plane possibly hosting a solution of the lens
        equation, for the given source position and lens model.

        :param sourcePos_x: source position in units of angle
        :param sourcePos_y: source position in units of angle
        :param kwargs_lens: lens model parameters as keyword arguments
        :param min_distance: minimum separation to consider for two images in units of
            angle
        :param search_window: window size to be considered by the solver. Will not find
            image position outside this window
        :param verbose: bool, if True, prints some useful information for the user
        :param x_center: float, center of the window to search for point sources
        :param y_center: float, center of the window to search for point sources
        :returns: (approximate) angular position of (multiple) images ra_pos, dec_pos in
            units of angles, related ray-traced source displacements and pixel width
        :raises: AttributeError, KeyError
        """
        kwargs_lens = self.lensModel.set_static(kwargs_lens)
        # compute number of pixels to cover the search window with the required min_distance
        numPix = int(round(search_window / min_distance) + 0.5)
        x_grid, y_grid = util.make_grid(numPix, min_distance)
        x_grid += x_center
        y_grid += y_center
        # ray-shoot to find the relative distance to the required source position for each grid point
        x_mapped, y_mapped = self.lensModel.ray_shooting(x_grid, y_grid, kwargs_lens)
        absmapped = util.displaceAbs(x_mapped, y_mapped, sourcePos_x, sourcePos_y)
        # select minima in the grid points and select grid points that do not deviate more than the
        # width of the grid point to a solution of the lens equation
        x_mins, y_mins, delta_map = util.local_minima_2d(absmapped, x_grid, y_grid)
        # pixel width
        pixel_width = x_grid[1] - x_grid[0]

        return x_mins, y_mins, delta_map, pixel_width

    def image_position_analytical(
        self,
        x,
        y,
        kwargs_lens,
        arrival_time_sort=True,
        magnification_limit=None,
        **kwargs_solver,
    ):
        """Solves the lens equation. Only supports EPL-like (plus shear) models. Uses a
        specialized recipe that solves a one-dimensional lens equation that is easier
        and more reliable to solve than the usual two-dimensional lens equation.

        :param x: source position in units of angle, an array of positions is also supported.
        :param y: source position in units of angle, an array of positions is also supported.
        :param kwargs_lens: lens model parameters as keyword arguments
        :param arrival_time_sort: bool, if True, sorts image position in arrival time (first arrival photon first listed)
        :param magnification_limit: None or float, if set will only return image positions that have an
         abs(magnification) larger than this number
        :param kwargs_solver: additional kwargs to be supplied to the solver. Particularly relevant are Nmeas and Nmeas_extra
        :returns: (exact) angular position of (multiple) images ra_pos, dec_pos in units of angle
         Note: in contrast to the other solvers, generally the (heavily demagnified) central image will also be included, so
         setting a a proper magnification_limit is more important. To get similar behaviour, a limit of 1e-1 is acceptable
        """
        lens_model_list = list(self.lensModel.lens_model_list)
        if lens_model_list not in (
            ["SIE", "SHEAR"],
            ["SIE"],
            ["EPL_NUMBA", "SHEAR"],
            ["EPL_NUMBA"],
            ["EPL", "SHEAR"],
            ["EPL"],
        ):
            raise ValueError(
                "Only SIE, EPL, EPL_NUMBA (+shear) supported in the analytical solver for now."
            )

        x_mins, y_mins = solve_lenseq_pemd((x, y), kwargs_lens, **kwargs_solver)
        if arrival_time_sort:
            x_mins, y_mins = self.sort_arrival_times(x_mins, y_mins, kwargs_lens)
        if magnification_limit is not None:
            mag = np.abs(self.lensModel.magnification(x_mins, y_mins, kwargs_lens))
            x_mins = x_mins[mag >= magnification_limit]
            y_mins = y_mins[mag >= magnification_limit]
        return x_mins, y_mins

    def image_position_from_source(
        self, sourcePos_x, sourcePos_y, kwargs_lens, solver="lenstronomy", **kwargs
    ):
        """Solves the lens equation, i.e. finds the image positions in the lens plane
        that are mapped to a given source position.

        :param sourcePos_x: source position in units of angle
        :param sourcePos_y: source position in units of angle
        :param kwargs_lens: lens model parameters as keyword arguments
        :param solver: which solver to use, can be 'lenstronomy' (default), 'analytical'
            or 'stochastic'.
        :param kwargs: Any additional kwargs are passed to the chosen solver, see the
            documentation of image_position_lenstronomy, image_position_analytical and
            image_position_stochastic
        :returns: (exact) angular position of (multiple) images ra_pos, dec_pos in units
            of angle
        """
        if solver == "lenstronomy":
            return self.image_position_lenstronomy(
                sourcePos_x, sourcePos_y, kwargs_lens, **kwargs
            )
        if solver == "analytical":
            return self.image_position_analytical(
                sourcePos_x, sourcePos_y, kwargs_lens, **kwargs
            )
        if solver == "stochastic":
            return self.image_position_stochastic(
                sourcePos_x, sourcePos_y, kwargs_lens, **kwargs
            )
        raise ValueError(f"{solver} is not a valid solver.")

    def image_position_lenstronomy(
        self,
        sourcePos_x,
        sourcePos_y,
        kwargs_lens,
        min_distance=0.1,
        search_window=10,
        precision_limit=10 ** (-10),
        num_iter_max=100,
        arrival_time_sort=True,
        initial_guess_cut=True,
        verbose=False,
        x_center=0,
        y_center=0,
        num_random=0,
        non_linear=False,
        magnification_limit=None,
    ):
        """Finds image position  given source position and lens model. The solver first
        samples does a grid search in the lens plane, and the grid points that are
        closest to the supplied source position are fed to a specialized gradient-based
        root finder that finds the exact solutions. Works with all lens models.

        :param sourcePos_x: source position in units of angle
        :param sourcePos_y: source position in units of angle
        :param kwargs_lens: lens model parameters as keyword arguments
        :param min_distance: minimum separation to consider for two images in units of
            angle
        :param search_window: window size to be considered by the solver. Will not find
            image position outside this window
        :param precision_limit: required precision in the lens equation solver (in units
            of angle in the source plane).
        :param num_iter_max: maximum iteration of lens-source mapping conducted by
            solver to match the required precision
        :param arrival_time_sort: bool, if True, sorts image position in arrival time
            (first arrival photon first listed)
        :param initial_guess_cut: bool, if True, cuts initial local minima selected by
            the grid search based on distance criteria from the source position
        :param verbose: bool, if True, prints some useful information for the user
        :param x_center: float, center of the window to search for point sources
        :param y_center: float, center of the window to search for point sources
        :param num_random: int, number of random positions within the search window to
            be added to be starting positions for the gradient decent solver
        :param non_linear: bool, if True applies a non-linear solver not dependent on
            Hessian computation
        :param magnification_limit: None or float, if set will only return image
            positions that have an abs(magnification) larger than this number
        :returns: (exact) angular position of (multiple) images ra_pos, dec_pos in units
            of angle
        :raises: AttributeError, KeyError
        """
        # find pixels in the image plane possibly hosting a solution of the lens equation, related source distances and
        # pixel width
        x_mins, y_mins, delta_map, pixel_width = self.candidate_solutions(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance,
            search_window,
            verbose,
            x_center,
            y_center,
        )
        if verbose:
            print(
                "There are %s regions identified that could contain a solution of the lens equation with"
                "coordinates %s and %s " % (len(x_mins), x_mins, y_mins)
            )
        if len(x_mins) < 1:
            return x_mins, y_mins
        if initial_guess_cut:
            mag = np.abs(self.lensModel.magnification(x_mins, y_mins, kwargs_lens))
            mag[mag < 1] = 1
            x_mins = x_mins[delta_map <= min_distance * mag * 5]
            y_mins = y_mins[delta_map <= min_distance * mag * 5]
            if verbose:
                print(
                    "The number of regions that meet the plausibility criteria are %s"
                    % len(x_mins)
                )
        x_mins = np.append(
            x_mins,
            np.random.uniform(
                low=-search_window / 2 + x_center,
                high=search_window / 2 + x_center,
                size=num_random,
            ),
        )
        y_mins = np.append(
            y_mins,
            np.random.uniform(
                low=-search_window / 2 + y_center,
                high=search_window / 2 + y_center,
                size=num_random,
            ),
        )
        # iterative solving of the lens equation for the selected grid points
        # print("Candidates:", x_mins.shape, y_mins.shape)
        x_mins, y_mins, solver_precision = self._find_gradient_decent(
            x_mins,
            y_mins,
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            precision_limit,
            num_iter_max,
            verbose=verbose,
            min_distance=min_distance,
            non_linear=non_linear,
        )
        # only select iterative results that match the precision limit
        x_mins = x_mins[solver_precision <= precision_limit]
        y_mins = y_mins[solver_precision <= precision_limit]
        # find redundant solutions within the min_distance criterion
        x_mins, y_mins = image_util.findOverlap(x_mins, y_mins, min_distance)
        if arrival_time_sort:
            x_mins, y_mins = self.sort_arrival_times(x_mins, y_mins, kwargs_lens)
        if magnification_limit is not None:
            mag = np.abs(self.lensModel.magnification(x_mins, y_mins, kwargs_lens))
            x_mins = x_mins[mag >= magnification_limit]
            y_mins = y_mins[mag >= magnification_limit]
        self.lensModel.set_dynamic()
        return x_mins, y_mins

    def _find_gradient_decent(
        self,
        x_min,
        y_min,
        sourcePos_x,
        sourcePos_y,
        kwargs_lens,
        precision_limit=10 ** (-10),
        num_iter_max=200,
        verbose=False,
        min_distance=0.01,
        non_linear=False,
    ):
        """Given a 'good guess' of a solution of the lens equation (expected image
        position given a fixed source position) this routine iteratively performs a ray-
        tracing with second order correction (effectively gradient decent) to find a
        precise solution to the lens equation.

        :param x_min: np.array, list of 'good guess' solutions of the lens equation
        :param y_min: np.array, list of 'good guess' solutions of the lens equation
        :param sourcePos_x: source position for which to solve the lens equation
        :param sourcePos_y: source position for which to solve the lens equation
        :param kwargs_lens: keyword argument list of the lens model
        :param precision_limit: float, required match in the solution in the source
            plane
        :param num_iter_max: int, maximum number of iterations before the algorithm
            stops
        :param verbose: bool, if True inserts print statements about the behavior of the
            solver
        :param min_distance: maximum correction applied per step (to avoid over-shooting
            in unstable regions)
        :param non_linear: bool, if True, uses scipy.miminize instead of the directly
            implemented gradient decent approach.
        :return: x_position array, y_position array, error in the source plane array
        """
        num_candidates = len(x_min)
        x_mins = np.zeros(num_candidates)
        y_mins = np.zeros(num_candidates)
        solver_precision = np.zeros(num_candidates)
        for i in range(len(x_min)):
            x_guess, y_guess, delta, l = self._solve_single_proposal(
                x_min[i],
                y_min[i],
                sourcePos_x,
                sourcePos_y,
                kwargs_lens,
                precision_limit,
                num_iter_max,
                max_step=min_distance,
                non_linear=non_linear,
            )
            if verbose:
                print(
                    "Solution found for region %s with required precision at iteration %s"
                    % (i, l)
                )
            x_mins[i] = x_guess
            y_mins[i] = y_guess
            solver_precision[i] = delta
        return x_mins, y_mins, solver_precision

    def _solve_single_proposal(
        self,
        x_guess,
        y_guess,
        source_x,
        source_y,
        kwargs_lens,
        precision_limit,
        num_iter_max,
        max_step,
        non_linear=False,
    ):
        """Gradient decent solution of a single proposed starting point (close to a true
        solution)

        :param x_guess: starting guess position in the image plane
        :param y_guess: starting guess position in the image plane
        :param source_x: source position to solve for in the image plane
        :param source_y: source position to solve for in the image plane
        :param kwargs_lens: keyword argument list of the lens model
        :param precision_limit: float, required match in the solution in the source
            plane
        :param num_iter_max: int, maximum number of iterations before the algorithm
            stops
        :param max_step: maximum correction applied per step (to avoid over-shooting in
            instable regions)
        :param non_linear: bool, if True, uses scipy.miminize instead of the directly
            implemented gradient decent approach.
        :return: x_position, y_position, error in the source plane, steps required (for
            gradient decent)
        """
        l = 0
        if non_linear:
            xinitial = np.array([x_guess, y_guess])
            result = minimize(
                self._root,
                xinitial,
                args=(kwargs_lens, source_x, source_y),
                tol=precision_limit**2,
                method="Nelder-Mead",
            )
            delta = self._root(result.x, kwargs_lens, source_x, source_y)
            x_guess, y_guess = result.x[0], result.x[1]

        else:
            x_mapped, y_mapped = self.lensModel.ray_shooting(
                x_guess, y_guess, kwargs_lens
            )
            delta = np.sqrt((x_mapped - source_x) ** 2 + (y_mapped - source_y) ** 2)

            while delta > precision_limit and l < num_iter_max:
                x_mapped, y_mapped = self.lensModel.ray_shooting(
                    x_guess, y_guess, kwargs_lens
                )
                delta = np.sqrt((x_mapped - source_x) ** 2 + (y_mapped - source_y) ** 2)
                f_xx, f_xy, f_yx, f_yy = self.lensModel.hessian(
                    x_guess, y_guess, kwargs_lens
                )
                DistMatrix = np.array([[1 - f_yy, f_yx], [f_xy, 1 - f_xx]])
                det = (1 - f_xx) * (1 - f_yy) - f_xy * f_yx
                deltaVec = np.array([x_mapped - source_x, y_mapped - source_y])
                image_plane_vector = DistMatrix.dot(deltaVec) / det
                dist = np.sqrt(image_plane_vector[0] ** 2 + image_plane_vector[1] ** 2)
                if dist > max_step:
                    image_plane_vector *= max_step / dist
                x_guess, y_guess, delta, l = self._gradient_step(
                    x_guess,
                    y_guess,
                    source_x,
                    source_y,
                    delta,
                    image_plane_vector,
                    kwargs_lens,
                    l,
                    num_iter_max,
                )
        return x_guess, y_guess, delta, l

    def _gradient_step(
        self,
        x_guess,
        y_guess,
        source_x,
        source_y,
        delta_init,
        image_plane_vector,
        kwargs_lens,
        iter_num,
        num_iter_max,
    ):
        """

        :param x_guess: float, current best fit solution in the image plane
        :param y_guess: float, current best fit solution in the image plane
        :param source_x: float, source position to be matched
        :param source_y: float, source position ot be matched
        :param delta_init: current precision in the source plane of the mapped solution
        :param image_plane_vector: correction vector in the image plane based on the Hessian operator and the deviation
         in the source plane
        :param kwargs_lens: lens model keyword argument list
        :param iter_num: int, current iteration number
        :param num_iter_max: int, maximum iteration number before aborting the process
        :return: updated image position in x, updated image position in y, updated precision in the source plane,
         total iterations done after this call
        """
        x_new = x_guess - image_plane_vector[0]
        y_new = y_guess - image_plane_vector[1]
        x_mapped, y_mapped = self.lensModel.ray_shooting(x_new, y_new, kwargs_lens)
        delta_new = np.sqrt((x_mapped - source_x) ** 2 + (y_mapped - source_y) ** 2)
        iter_num += 1
        if delta_new > delta_init:
            if num_iter_max < iter_num:
                return x_guess, y_guess, delta_init, iter_num
            else:
                # if the new proposal is worse than the previous one, it randomly draws a new proposal in a different
                # direction and tries again
                image_plane_vector[0] *= np.random.normal(loc=0, scale=0.5)
                image_plane_vector[1] *= np.random.normal(loc=0, scale=0.5)
                return self._gradient_step(
                    x_guess,
                    y_guess,
                    source_x,
                    source_y,
                    delta_init,
                    image_plane_vector,
                    kwargs_lens,
                    iter_num,
                    num_iter_max,
                )
        else:
            return x_new, y_new, delta_new, iter_num

    def findBrightImage(
        self,
        sourcePos_x,
        sourcePos_y,
        kwargs_lens,
        numImages=4,
        min_distance=0.01,
        search_window=5,
        precision_limit=10 ** (-10),
        num_iter_max=10,
        arrival_time_sort=True,
        x_center=0,
        y_center=0,
        num_random=0,
        non_linear=False,
        magnification_limit=None,
        initial_guess_cut=True,
        verbose=False,
    ):
        """

        :param sourcePos_x: source position in units of angle
        :param sourcePos_y: source position in units of angle
        :param kwargs_lens: lens model parameters as keyword arguments
        :param min_distance: minimum separation to consider for two images in units of angle
        :param search_window: window size to be considered by the solver. Will not find image position outside this window
        :param precision_limit: required precision in the lens equation solver (in units of angle in the source plane).
        :param num_iter_max: maximum iteration of lens-source mapping conducted by solver to match the required precision
        :param arrival_time_sort: bool, if True, sorts image position in arrival time (first arrival photon first listed)
        :param initial_guess_cut: bool, if True, cuts initial local minima selected by the grid search based on distance criteria from the source position
        :param verbose: bool, if True, prints some useful information for the user
        :param x_center: float, center of the window to search for point sources
        :param y_center: float, center of the window to search for point sources
        :param num_random: int, number of random positions within the search window to be added to be starting
         positions for the gradient decent solver
        :param non_linear: bool, if True applies a non-linear solver not dependent on Hessian computation
        :param magnification_limit: None or float, if set will only return image positions that have an
         abs(magnification) larger than this number
        :returns: (exact) angular position of (multiple) images ra_pos, dec_pos in units of angle
        """

        x_mins, y_mins = self.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=precision_limit,
            num_iter_max=num_iter_max,
            arrival_time_sort=arrival_time_sort,
            initial_guess_cut=initial_guess_cut,
            verbose=verbose,
            x_center=x_center,
            y_center=y_center,
            num_random=num_random,
            non_linear=non_linear,
            magnification_limit=magnification_limit,
        )
        mag_list = []
        for i in range(len(x_mins)):
            mag = self.lensModel.magnification(x_mins[i], y_mins[i], kwargs_lens)
            mag_list.append(abs(mag))
        mag_list = np.array(mag_list)
        x_mins_sorted = util.selectBest(x_mins, mag_list, numImages)
        y_mins_sorted = util.selectBest(y_mins, mag_list, numImages)
        if arrival_time_sort:
            x_mins_sorted, y_mins_sorted = self.sort_arrival_times(
                x_mins_sorted, y_mins_sorted, kwargs_lens
            )
        return x_mins_sorted, y_mins_sorted

    def sort_arrival_times(self, x_mins, y_mins, kwargs_lens):
        """Sort arrival times (fermat potential) of image positions in increasing order
        of light travel time.

        :param x_mins: ra position of images
        :param y_mins: dec position of images
        :param kwargs_lens: keyword arguments of lens model
        :return: sorted lists of x_mins and y_mins
        """

        if hasattr(self.lensModel, "_no_potential"):
            raise Exception(
                "Instance of lensModel passed to this class does not compute the lensing potential, "
                "and therefore cannot compute time delays."
            )

        if len(x_mins) <= 1:
            return x_mins, y_mins
        if self.lensModel.multi_plane:
            arrival_time = self.lensModel.arrival_time(x_mins, y_mins, kwargs_lens)
        else:
            fermat_pot = self.lensModel.fermat_potential(x_mins, y_mins, kwargs_lens)
            arrival_time = fermat_pot
        idx = np.argsort(arrival_time)
        x_mins = np.array(x_mins)[idx]
        y_mins = np.array(y_mins)[idx]
        return x_mins, y_mins
